# sage.doctest: needs sage.combinat sage.modules
r"""
Rigged configurations of `\mathcal{B}(\infty)`

AUTHORS:

- Travis Scrimshaw (2013-04-16): initial version
"""

# ****************************************************************************
#       Copyright (C) 2013 Travis Scrimshaw <tscrim@ucdavis.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.misc.lazy_attribute import lazy_attribute
from sage.structure.unique_representation import UniqueRepresentation
from sage.structure.parent import Parent
from sage.categories.highest_weight_crystals import HighestWeightCrystals
from sage.categories.homset import Hom
from sage.combinat.root_system.cartan_type import CartanType
from sage.combinat.rigged_configurations.rigged_configuration_element import (
     RiggedConfigurationElement, RCNonSimplyLacedElement)
from sage.combinat.rigged_configurations.rigged_configurations import RiggedConfigurations
from sage.combinat.rigged_configurations.rigged_partition import RiggedPartition

# Note on implementation, this class is used for simply-laced types only


class InfinityCrystalOfRiggedConfigurations(UniqueRepresentation, Parent):
    r"""
    Rigged configuration model for `\mathcal{B}(\infty)`.

    The crystal is generated by the empty rigged configuration with the same
    crystal structure given by the :class:`highest weight model
    <sage.combinat.rigged_configurations.rc_crystal.CrystalOfRiggedConfigurations>`
    except we remove the condition that the resulting rigged configuration
    needs to be valid when applying `f_a`.

    INPUT:

    - ``cartan_type`` -- a Cartan type

    EXAMPLES:

    For simplicity, we display all of the rigged configurations
    horizontally::

        sage: RiggedConfigurations.options(display='horizontal')

    We begin with a simply-laced finite type::

        sage: RC = crystals.infinity.RiggedConfigurations(['A', 3]); RC
        The infinity crystal of rigged configurations of type ['A', 3]

        sage: RC.options(display='horizontal')

        sage: mg = RC.highest_weight_vector(); mg
        (/)  (/)  (/)
        sage: elt = mg.f_string([2,1,3,2]); elt
        0[ ]0   -2[ ]-1   0[ ]0
                -2[ ]-1
        sage: elt.e(1)
        sage: elt.e(3)
        sage: mg.f_string([2,1,3,2]).e(2)
        -1[ ]-1  0[ ]1  -1[ ]-1
        sage: mg.f_string([2,3,2,1,3,2])
        0[ ]0  -3[ ][ ]-1  -1[ ][ ]-1
               -2[ ]-1

    Next we consider a non-simply-laced finite type::

        sage: RC = crystals.infinity.RiggedConfigurations(['C', 3])
        sage: mg = RC.highest_weight_vector()
        sage: mg.f_string([2,1,3,2])
        0[ ]0   -1[ ]0    0[ ]0
                -1[ ]-1
        sage: mg.f_string([2,3,2,1,3,2])
        0[ ]-1   -1[ ][ ]-1   -1[ ][ ]0
                 -1[ ]0

    We can construct rigged configurations using a diagram folding of
    a simply-laced type. This yields an equivalent but distinct crystal::

        sage: vct = CartanType(['C', 3]).as_folding()
        sage: VRC = crystals.infinity.RiggedConfigurations(vct)
        sage: mg = VRC.highest_weight_vector()
        sage: mg.f_string([2,1,3,2])
        0[ ]0   -2[ ]-1   0[ ]0
                -2[ ]-1
        sage: mg.f_string([2,3,2,1,3,2])
        -1[ ]-1  -2[ ][ ][ ]-1  -1[ ][ ]0

        sage: G = RC.subcrystal(max_depth=5).digraph()
        sage: VG = VRC.subcrystal(max_depth=5).digraph()
        sage: G.is_isomorphic(VG, edge_labels=True)
        True

    We can also construct `B(\infty)` using rigged configurations in
    affine types::

        sage: RC = crystals.infinity.RiggedConfigurations(['A', 3, 1])
        sage: mg = RC.highest_weight_vector()
        sage: mg.f_string([0,1,2,3,0,1,3])
        -1[ ]0  -1[ ]-1  1[ ]1  -1[ ][ ]-1
        -1[ ]0  -1[ ]-1

        sage: RC = crystals.infinity.RiggedConfigurations(['C', 3, 1])
        sage: mg = RC.highest_weight_vector()
        sage: mg.f_string([1,2,3,0,1,2,3,3,0])
        -2[ ][ ]-1   0[ ]1   0[ ]0    -4[ ][ ][ ]-2
                     0[ ]0   0[ ]-1

        sage: RC = crystals.infinity.RiggedConfigurations(['A', 6, 2])
        sage: mg = RC.highest_weight_vector()
        sage: mg.f_string([1,2,3,0,1,2,3,3,0])
        0[ ]-1   0[ ]1   0[ ]0    -4[ ][ ][ ]-2
        0[ ]-1   0[ ]1   0[ ]-1

    We reset the global options::

        sage: RiggedConfigurations.options._reset()
    """
    @staticmethod
    def __classcall_private__(cls, cartan_type):
        r"""
        Normalize the input arguments to ensure unique representation.

        EXAMPLES::

            sage: RC1 = crystals.infinity.RiggedConfigurations(CartanType(['A',3]))
            sage: RC2 = crystals.infinity.RiggedConfigurations(['A',3])
            sage: RC2 is RC1
            True
        """
        from sage.combinat.root_system.type_folded import CartanTypeFolded
        if isinstance(cartan_type, CartanTypeFolded):
            return InfinityCrystalOfNonSimplyLacedRC(cartan_type)

        cartan_type = CartanType(cartan_type)
        return super().__classcall__(cls, cartan_type)

    def __init__(self, cartan_type):
        r"""
        Initialize ``self``.

        EXAMPLES::

            sage: RC = crystals.infinity.RiggedConfigurations(['A', 2])
            sage: TestSuite(RC).run()
            sage: RC = crystals.infinity.RiggedConfigurations(['A', 2, 1])
            sage: TestSuite(RC).run() # long time
            sage: RC = crystals.infinity.RiggedConfigurations(['C', 2])
            sage: TestSuite(RC).run() # long time
            sage: RC = crystals.infinity.RiggedConfigurations(['C', 2, 1])
            sage: TestSuite(RC).run() # long time
        """
        self._cartan_type = cartan_type
        Parent.__init__(self, category=HighestWeightCrystals().Infinite())
        self._rc_index = self._cartan_type.index_set()
        self._rc_index_inverse = {i: ii for ii, i in enumerate(self._rc_index)}
        # We store the Cartan matrix for the vacancy number
        # calculations for speed
        self._cartan_matrix = self._cartan_type.cartan_matrix()
        self.module_generators = (self.element_class(self, rigging_list=[[]]*cartan_type.rank()),)

    options = RiggedConfigurations.options

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: crystals.infinity.RiggedConfigurations(['A', 3])
            The infinity crystal of rigged configurations of type ['A', 3]
        """
        return "The infinity crystal of rigged configurations of type {}".format(self._cartan_type)

    def _element_constructor_(self, lst=None, **options):
        """
        Construct an element of ``self`` from ``lst``.

        EXAMPLES::

            sage: RC = crystals.infinity.RiggedConfigurations(['A', 3, 1])
            sage: ascii_art(RC(partition_list=[[1,1]]*4, rigging_list=[[1,1], [0,0], [0,0], [-1,-1]]))
            0[ ]1  0[ ]0  0[ ]0  0[ ]-1
            0[ ]1  0[ ]0  0[ ]0  0[ ]-1

            sage: RC = crystals.infinity.RiggedConfigurations(['C', 3])
            sage: ascii_art(RC(partition_list=[[1],[1,1],[1]], rigging_list=[[0],[0,-1],[0]]))
            0[ ]0  -1[ ]0   0[ ]0
                   -1[ ]-1

        TESTS:

        Check that :issue:`17054` is fixed::

            sage: RC = RiggedConfigurations(['A',2,1], [[1,1]]*4 + [[2,1]]*4)
            sage: B = crystals.infinity.RiggedConfigurations(['A',2])
            sage: x = RC().f_string([2,2,1,1,2,1,2,1])
            sage: ascii_art(x)
            0[ ][ ][ ][ ]-4  0[ ][ ][ ][ ]0
            sage: ascii_art(B(x))
            -4[ ][ ][ ][ ]-4  -4[ ][ ][ ][ ]0
            sage: x == RC().f_string([2,2,1,1,2,1,2,1])
            True
        """
        if isinstance(lst, RiggedConfigurationElement):
            lst = [p._clone() for p in lst]  # Make a deep copy
        elif isinstance(lst, list) and bool(lst) and isinstance(lst[0], RiggedPartition):
            lst = [p._clone() for p in lst]  # Make a deep copy
        return self.element_class(self, lst, **options)

    def _coerce_map_from_(self, P):
        """
        Return ``True`` or the coerce map from ``P`` if a map exists.

        EXAMPLES::

            sage: T = crystals.infinity.Tableaux(['A',3])
            sage: RC = crystals.infinity.RiggedConfigurations(['A',3])
            sage: RC._coerce_map_from_(T)
            Crystal Isomorphism morphism:
              From: The infinity crystal of tableaux of type ['A', 3]
              To:   The infinity crystal of rigged configurations of type ['A', 3]
        """
        if self.cartan_type().is_finite():
            from sage.combinat.crystals.infinity_crystals import InfinityCrystalOfTableaux
            if (isinstance(P, InfinityCrystalOfTableaux)
                and self.cartan_type().is_simply_laced()):
                from sage.combinat.rigged_configurations.bij_infinity import FromTableauIsomorphism
                return FromTableauIsomorphism(Hom(P, self))
        return super()._coerce_map_from_(P)

    def _calc_vacancy_number(self, partitions, a, i, **options):
        r"""
        Calculate the vacancy number `p_i^{(a)}(\nu)` in ``self``.

        This assumes that `\gamma_a = 1` for all `a` and `(\alpha_a \mid
        \alpha_b ) = A_{ab}`.

        INPUT:

        - ``partitions`` -- the list of rigged partitions we are using

        - ``a`` -- the rigged partition index

        - ``i`` -- the row length

        TESTS::

            sage: RC = crystals.infinity.RiggedConfigurations(['A', 4])
            sage: elt = RC(partition_list=[[1], [1], [], []])
            sage: RC._calc_vacancy_number(elt.nu(), 0, 1)
            -1
        """
        if i == float('inf'):
            return -sum(self._cartan_matrix[a, b] * sum(nu)
                        for b, nu in enumerate(partitions))

        return -sum(self._cartan_matrix[a, b] * nu.get_num_cells_to_column(i)
                    for b, nu in enumerate(partitions))

    # FIXME: Remove this method!!!
    def weight_lattice_realization(self):
        """
        Return the weight lattice realization used to express the weights
        of elements in ``self``.

        EXAMPLES::

            sage: RC = crystals.infinity.RiggedConfigurations(['A', 2, 1])
            sage: RC.weight_lattice_realization()
            Extended weight lattice of the Root system of type ['A', 2, 1]
        """
        R = self._cartan_type.root_system()
        if self._cartan_type.is_affine():
            return R.weight_lattice(extended=True)
        if self._cartan_type.is_finite() and R.ambient_space():
            return R.ambient_space()
        return R.weight_lattice()

    class Element(RiggedConfigurationElement):
        r"""
        A rigged configuration in `\mathcal{B}(\infty)` in simply-laced types.

        TESTS::

            sage: RC = crystals.infinity.RiggedConfigurations(['A', 3, 1])
            sage: elt = RC(partition_list=[[1,1]]*4, rigging_list=[[1,1], [0,0], [0,0], [-1,-1]])
            sage: TestSuite(elt).run()
        """

        def weight(self):
            """
            Return the weight of ``self``.

            EXAMPLES::

                sage: RC = crystals.infinity.RiggedConfigurations(['A', 3, 1])
                sage: elt = RC(partition_list=[[1,1]]*4, rigging_list=[[1,1], [0,0], [0,0], [-1,-1]])
                sage: elt.weight()
                -2*delta
            """
            P = self.parent().weight_lattice_realization()
            alpha = list(P.simple_roots())
            return -sum(sum(x) * alpha[i] for i, x in enumerate(self))


class InfinityCrystalOfNonSimplyLacedRC(InfinityCrystalOfRiggedConfigurations):
    r"""
    Rigged configurations for `\mathcal{B}(\infty)` in non-simply-laced types.
    """

    def __init__(self, vct):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: vct = CartanType(['C', 2]).as_folding()
            sage: RC = crystals.infinity.RiggedConfigurations(vct)
            sage: TestSuite(RC).run() # long time
            sage: vct = CartanType(['C', 2, 1]).as_folding()
            sage: RC = crystals.infinity.RiggedConfigurations(vct)
            sage: TestSuite(RC).run() # long time
        """
        self._folded_ct = vct
        InfinityCrystalOfRiggedConfigurations.__init__(self, vct._cartan_type)

    def _coerce_map_from_(self, P):
        """
        Return ``True`` or the coerce map from ``P`` if a map exists.

        EXAMPLES::

            sage: T = crystals.infinity.Tableaux(['C',3])
            sage: vct = CartanType(['C',3]).as_folding()
            sage: RC = crystals.infinity.RiggedConfigurations(vct)
            sage: RC._coerce_map_from_(T)
            Crystal Isomorphism morphism:
              From: The infinity crystal of tableaux of type ['C', 3]
              To:   The infinity crystal of rigged configurations of type ['C', 3]
        """
        if self.cartan_type().is_finite():
            from sage.combinat.crystals.infinity_crystals import InfinityCrystalOfTableaux
            if isinstance(P, InfinityCrystalOfTableaux):
                from sage.combinat.rigged_configurations.bij_infinity import FromTableauIsomorphism
                return FromTableauIsomorphism(Hom(P, self))
        return super()._coerce_map_from_(P)

    def _calc_vacancy_number(self, partitions, a, i):
        r"""
        Calculate the vacancy number `p_i^{(a)}(\nu)` in ``self``.

        INPUT:

        - ``partitions`` -- the list of rigged partitions we are using

        - ``a`` -- the rigged partition index

        - ``i`` -- the row length

        TESTS::

            sage: La = RootSystem(['C', 2]).weight_lattice().fundamental_weights()
            sage: vct = CartanType(['C', 2]).as_folding()
            sage: RC = crystals.RiggedConfigurations(vct, La[1])
            sage: elt = RC(partition_list=[[1], [1]])
            sage: RC._calc_vacancy_number(elt.nu(), 0, 1)
            0
            sage: RC._calc_vacancy_number(elt.nu(), 1, 1)
            -1
        """
        I = self.index_set()
        ia = I[a]
        vac_num = 0

        if i == float('inf'):
            return -sum(self._cartan_matrix[a, b] * sum(nu)
                        for b, nu in enumerate(partitions))

        gamma = self._folded_ct.scaling_factors()
        g = gamma[ia]
        for b in range(self._cartan_matrix.ncols()):
            ib = I[b]
            q = partitions[b].get_num_cells_to_column(g * i, gamma[ib])
            vac_num -= self._cartan_matrix[a, b] * q // gamma[ib]

        return vac_num

    @lazy_attribute
    def virtual(self):
        """
        Return the corresponding virtual crystal.

        EXAMPLES::

            sage: vct = CartanType(['C', 3]).as_folding()
            sage: RC = crystals.infinity.RiggedConfigurations(vct)
            sage: RC
            The infinity crystal of rigged configurations of type ['C', 3]
            sage: RC.virtual
            The infinity crystal of rigged configurations of type ['A', 5]
        """
        return InfinityCrystalOfRiggedConfigurations(self._folded_ct._folding)

    def to_virtual(self, rc):
        """
        Convert ``rc`` into a rigged configuration in the virtual crystal.

        INPUT:

        - ``rc`` -- a rigged configuration element

        EXAMPLES::

            sage: vct = CartanType(['C', 2]).as_folding()
            sage: RC = crystals.infinity.RiggedConfigurations(vct)
            sage: mg = RC.highest_weight_vector()
            sage: elt = mg.f_string([1,2,2,1,1]); elt
            <BLANKLINE>
            -3[ ][ ][ ]-2
            <BLANKLINE>
            -1[ ][ ]0
            <BLANKLINE>
            sage: velt = RC.to_virtual(elt); velt
            <BLANKLINE>
            -3[ ][ ][ ]-2
            <BLANKLINE>
            -2[ ][ ][ ][ ]0
            <BLANKLINE>
            -3[ ][ ][ ]-2
            <BLANKLINE>
            sage: velt.parent()
            The infinity crystal of rigged configurations of type ['A', 3]
        """
        gamma = [int(f) for f in self._folded_ct.scaling_factors()]
        sigma = self._folded_ct._orbit
        n = self._folded_ct._folding.rank()
        vindex = self._folded_ct._folding.index_set()
        partitions = [None] * n
        riggings = [None] * n
        for a, rp in enumerate(rc):
            for i in sigma[a]:
                k = vindex.index(i)
                partitions[k] = [row_len * gamma[a] for row_len in rp._list]
                riggings[k] = [rig_val * gamma[a] for rig_val in rp.rigging]
        return self.virtual.element_class(self.virtual, partition_list=partitions,
                                          rigging_list=riggings)

    def from_virtual(self, vrc):
        """
        Convert ``vrc`` in the virtual crystal into a rigged configuration of
        the original Cartan type.

        INPUT:

        - ``vrc`` -- a virtual rigged configuration

        EXAMPLES::

            sage: vct = CartanType(['C', 2]).as_folding()
            sage: RC = crystals.infinity.RiggedConfigurations(vct)
            sage: elt = RC(partition_list=[[3],[2]], rigging_list=[[-2],[0]])
            sage: vrc_elt = RC.to_virtual(elt)
            sage: ret = RC.from_virtual(vrc_elt); ret
            <BLANKLINE>
            -3[ ][ ][ ]-2
            <BLANKLINE>
            -1[ ][ ]0
            <BLANKLINE>
            sage: ret == elt
            True
        """
        gamma = list(self._folded_ct.scaling_factors())  # map(int, self._folded_ct.scaling_factors())
        sigma = self._folded_ct._orbit
        n = self._cartan_type.rank()
        partitions = [None] * n
        riggings = [None] * n
        vindex = self._folded_ct._folding.index_set()
        # TODO: Handle special cases for A^{(2)} even and its dual?
        for a in range(n):
            index = vindex.index(sigma[a][0])
            partitions[a] = [row_len // gamma[a] for row_len in vrc[index]._list]
            riggings[a] = [rig_val / gamma[a] for rig_val in vrc[index].rigging]
        return self.element_class(self, partition_list=partitions,
                                  rigging_list=riggings)

    class Element(RCNonSimplyLacedElement):
        r"""
        A rigged configuration in `\mathcal{B}(\infty)` in
        non-simply-laced types.

        TESTS::

            sage: vct = CartanType(['C', 3]).as_folding()
            sage: RC = crystals.infinity.RiggedConfigurations(vct)
            sage: elt = RC(partition_list=[[1],[1,1],[1]])
            sage: TestSuite(elt).run()
        """

        def weight(self):
            """
            Return the weight of ``self``.

            EXAMPLES::

                sage: vct = CartanType(['C', 3]).as_folding()
                sage: RC = crystals.infinity.RiggedConfigurations(vct)
                sage: elt = RC(partition_list=[[1],[1,1],[1]], rigging_list=[[0],[-1,-1],[0]])
                sage: elt.weight()
                (-1, -1, 0)

                sage: vct = CartanType(['F', 4, 1]).as_folding()
                sage: RC = crystals.infinity.RiggedConfigurations(vct)
                sage: mg = RC.highest_weight_vector()
                sage: elt = mg.f_string([1,0,3,4,2,2]); ascii_art(elt)
                -1[ ]-1  0[ ]1  -2[ ][ ]-2  0[ ]1  -1[ ]-1
                sage: wt = elt.weight(); wt
                -Lambda[0] + Lambda[1] - 2*Lambda[2] + 3*Lambda[3] - Lambda[4] - delta
                sage: al = RC.weight_lattice_realization().simple_roots()
                sage: wt == -(al[0] + al[1] + 2*al[2] + al[3] + al[4])
                True
            """
            P = self.parent().weight_lattice_realization()
            alpha = list(P.simple_roots())
            return -sum(sum(x) * alpha[i] for i, x in enumerate(self))



r"""
Quantum Matrix Coordinate Algebras

AUTHORS:

- Travis Scrimshaw (01-2016): initial version
"""

##############################################################################
#       Copyright (C) 2016 Travis Scrimshaw <tscrimsh at umn.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
##############################################################################

from sage.misc.cachefunc import cached_method
from sage.misc.lazy_attribute import lazy_attribute
from sage.sets.family import Family
from sage.categories.algebras import Algebras
from sage.categories.bialgebras import Bialgebras
from sage.categories.hopf_algebras import HopfAlgebras
from sage.combinat.free_module import CombinatorialFreeModule
from sage.monoids.indexed_free_monoid import IndexedFreeAbelianMonoid
from sage.rings.polynomial.laurent_polynomial_ring import LaurentPolynomialRing
from sage.rings.integer_ring import ZZ


class QuantumMatrixCoordinateAlgebra_abstract(CombinatorialFreeModule):
    """
    Abstract base class for quantum coordinate algebras of a set
    of matrices.
    """
    @staticmethod
    def __classcall__(cls, q=None, bar=None, R=None, **kwds):
        """
        Normalize input to ensure a unique representation.

        EXAMPLES::

            sage: R.<q> = LaurentPolynomialRing(ZZ)
            sage: O1 = algebras.QuantumMatrixCoordinate(4)
            sage: O2 = algebras.QuantumMatrixCoordinate(4, 4, q=q)
            sage: O3 = algebras.QuantumMatrixCoordinate(4, R=ZZ)
            sage: O4 = algebras.QuantumMatrixCoordinate(4, R=R, q=q)
            sage: O1 is O2 and O2 is O3 and O3 is O4
            True
            sage: O5 = algebras.QuantumMatrixCoordinate(4, R=QQ)
            sage: O1 is O5
            False
        """
        if R is None:
            R = ZZ
        else:
            if q is not None:
                q = R(q)
        if q is None:
            q = LaurentPolynomialRing(R, 'q').gen()
        return super().__classcall__(cls,
                                     q=q, bar=bar, R=q.parent(), **kwds)

    def __init__(self, gp_indices, n, q, bar, R, category, indices_key=None):
        """
        Initialize ``self``.

        TESTS::

            sage: O = algebras.QuantumMatrixCoordinate(3, 2)
            sage: TestSuite(O).run()
        """
        self._n = n
        self._q = q
        if bar is None:
            def bar(x):
                return x.subs(q=~self._q)
        self._bar = bar
        if indices_key is None:
            indices = IndexedFreeAbelianMonoid(gp_indices)
        else:
            indices = IndexedFreeAbelianMonoid(gp_indices, sorting_key=indices_key)
        CombinatorialFreeModule.__init__(self, R, indices, category=category)

    def _repr_term(self, m) -> str:
        r"""
        Return a string representation of the term indexed by ``m``.

        EXAMPLES::

            sage: O = algebras.QuantumMatrixCoordinate(4)
            sage: I = O.indices()
            sage: x = I.an_element(); x
            F[(1, 1)]^2*F[(1, 2)]^2*F[(1, 3)]^3
            sage: O._repr_term(x)
            'x[1,1]^2*x[1,2]^2*x[1,3]^3'
            sage: O._repr_term(I.one())
            '1'
            sage: O.q() * O.one()
            q
        """
        S = m._sorted_items()
        if not S:
            return '1'

        def exp(e):
            return '^{}'.format(e) if e > 1 else ''
        return '*'.join(('x[{},{}]'.format(*k) if k != 'c' else 'c') + exp(e)
                        for k, e in m._sorted_items())

    def _latex_term(self, m) -> str:
        r"""
        Return a latex representation of the term indexed by ``m``.

        EXAMPLES::

            sage: O = algebras.QuantumMatrixCoordinate(4)
            sage: I = O.indices()
            sage: x = I.an_element(); x
            F[(1, 1)]^2*F[(1, 2)]^2*F[(1, 3)]^3
            sage: O._latex_term(x)
            'x_{1,1}^{2} x_{1,2}^{2} x_{1,3}^{3}'
            sage: O._latex_term(I.one())
            '1'
            sage: latex(O.q() * O.one())
            q
        """
        S = m._sorted_items()
        if not S:
            return '1'

        def exp(e):
            return '^{{{}}}'.format(e) if e > 1 else ''
        return ' '.join(('x_{{{},{}}}'.format(*k) if k != 'c' else 'c') + exp(e)
                        for k, e in m._sorted_items())

    def n(self):
        """
        Return the value `n`.

        EXAMPLES::

            sage: O = algebras.QuantumMatrixCoordinate(4)
            sage: O.n()
            4
            sage: O = algebras.QuantumMatrixCoordinate(4, 6)
            sage: O.n()
            6
        """
        return self._n

    def q(self):
        """
        Return the variable ``q``.

        EXAMPLES::

            sage: O = algebras.QuantumMatrixCoordinate(4)
            sage: O.q()
            q
            sage: O.q().parent()
            Univariate Laurent Polynomial Ring in q over Integer Ring
            sage: O.q().parent() is O.base_ring()
            True
        """
        return self._q

    @cached_method
    def one_basis(self):
        """
        Return the basis element indexing `1`.

        EXAMPLES::

            sage: O = algebras.QuantumMatrixCoordinate(4)
            sage: O.one_basis()
            1
            sage: O.one()
            1

        TESTS::

            sage: O = algebras.QuantumMatrixCoordinate(4)
            sage: O.one_basis() == O.indices().one()
            True
        """
        return self._indices.one()

    @cached_method
    def gens(self) -> tuple:
        r"""
        Return the generators of ``self`` as a tuple.

        EXAMPLES::

            sage: O = algebras.QuantumMatrixCoordinate(3)
            sage: O.gens()
            (x[1,1], x[1,2], x[1,3],
             x[2,1], x[2,2], x[2,3],
             x[3,1], x[3,2], x[3,3])
        """
        return tuple(self.algebra_generators())

    @cached_method
    def quantum_determinant(self):
        r"""
        Return the quantum determinant of ``self``.

        The quantum determinant is defined by

        .. MATH::

            \det_q = \sum_{\sigma \in S_n} (-q)^{\ell(\sigma)}
            x_{1, \sigma(1)} x_{2, \sigma(2)} \cdots x_{n, \sigma(n)}.

        EXAMPLES::

            sage: O = algebras.QuantumMatrixCoordinate(2)
            sage: O.quantum_determinant()
            x[1,1]*x[2,2] - q*x[1,2]*x[2,1]

        We verify that the quantum determinant is central::

            sage: for n in range(2,5):
            ....:     O = algebras.QuantumMatrixCoordinate(n)
            ....:     qdet = O.quantum_determinant()
            ....:     assert all(g * qdet == qdet * g for g in O.algebra_generators())

        We also verify that it is group-like::

            sage: for n in range(2,4):
            ....:     O = algebras.QuantumMatrixCoordinate(n)
            ....:     qdet = O.quantum_determinant()
            ....:     assert qdet.coproduct() == tensor([qdet, qdet])
        """
        if hasattr(self, '_m') and self._m != self._n:
            raise ValueError("undefined for non-square quantum matrices")
        from sage.combinat.permutation import Permutations
        q = self._q
        return self._from_dict({self._indices({(i, p(i)): 1 for i in range(1, self._n + 1)}):
                               (-q) ** p.length() for p in Permutations(self._n)})

    def product_on_basis(self, a, b):
        """
        Return the product of basis elements indexed by ``a`` and ``b``.

        EXAMPLES::

            sage: O = algebras.QuantumMatrixCoordinate(4)
            sage: x = O.algebra_generators()
            sage: b = x[1,4] * x[2,1] * x[3,4]  # indirect doctest
            sage: b * (b * b) == (b * b) * b
            True
            sage: p = prod(list(O.algebra_generators())[:10])
            sage: p * (p * p) == (p * p) * p  # long time
            True
            sage: x = O.an_element()
            sage: y = x^2 + x[4,4] * x[3,3] * x[1,2]
            sage: z = x[2,2] * x[1,4] * x[3,4] * x[1,1]
            sage: x * (y * z) == (x * y) * z
            True
        """
        al = a._sorted_items()
        bl = b._sorted_items()
        # Check for multiplication by 1
        if not al:
            return self.monomial(b)
        if not bl:
            return self.monomial(a)
        if al[-1][0] < bl[0][0]:  # Already in order
            return self.monomial(a * b)
        G = self._indices.monoid_generators()
        one = self.base_ring().one()
        q = self._q
        qi = q ** -1
        monomial = b
        coeff = one
        for pos in range(len(al) - 1, -1, -1):
            ax, ae = al[pos]
            for bx, be in bl:
                if ax[0] < bx[0]:
                    # In order, so nothing more to do
                    break
                if ax[0] == bx[0]:
                    if ax[1] > bx[1]:
                        # x_{it} x_{ij} = q^{-1} x_{ij} x_{it} if t < j
                        coeff *= qi ** (ae * be)
                    else:
                        # In order, so nothing more to do
                        break
                elif ax[1] == bx[1]:
                    # x_{sj} x_{ij} = q^{-1} x_{ij} x_{sj} if s > i
                    coeff *= qi ** (ae * be)
                elif ax[1] > bx[1]:  # By this point, we must have ax[0] > bx[0]
                    # x_{st} x_{ij} = x_{ij} x_{st} + (q^-1 - q) x_{it} x_{sj}
                    # if s > i, t > j

                    # By Lemma 2.7 (with fixed typo) in H. Zhang and R.B. Zhang:
                    # x_{st} x_{ij}^k = x_{ij}^k x_{st}
                    #                  + (q^{1-2k} - q) x_{ij}^{k-1} x_{it} x_{sj}
                    m1 = G[bx] ** be * G[ax]
                    m2 = G[bx] ** (be - 1) * G[(bx[0], ax[1])] * G[(ax[0], bx[1])]
                    ret = self._from_dict({m1: one, m2: (q ** (1 - 2 * be) - q)})
                    ml = monomial._sorted_items()
                    index = ml.index((bx, be))
                    a_key = self._indices(dict(al[:pos]))
                    bp_key = self._indices(dict(ml[:index])) * G[ax] ** (ae - 1)
                    return (self.monomial(a_key) *
                            self.monomial(bp_key) *
                            ret *
                            self.term(self._indices(dict(ml[index + 1:])),
                                      coeff))

                # Otherwise ax[1] > bx[1], but for this case they commute:
                # x_{st} x_{ij} = x_{ij} x_{st} if s > i, t < j
                # So there is nothing to do to coeff
            monomial *= G[ax] ** ae
        return self.term(monomial, coeff)

    @cached_method
    def _bar_on_basis(self, x):
        """
        Return the bar involution on the basis element indexed by ``x``.

        EXAMPLES::

            sage: O = algebras.QuantumMatrixCoordinate(4)
            sage: O._bar_on_basis(O._indices.an_element())
            (q^-16)*x[1,1]^2*x[1,2]^2*x[1,3]^3
        """
        ret = self.one()
        for k, e in reversed(x._sorted_items()):
            ret *= self.monomial(self._indices({k: e}))
        return ret

    def counit_on_basis(self, x):
        r"""
        Return the counit on the basis element indexed by ``x``.

        EXAMPLES::

            sage: O = algebras.QuantumMatrixCoordinate(4)
            sage: G = O.algebra_generators()
            sage: I = [1,2,3,4]
            sage: matrix([[G[i,j].counit() for i in I] for j in I])  # indirect doctest
            [1 0 0 0]
            [0 1 0 0]
            [0 0 1 0]
            [0 0 0 1]
        """
        if all(t == 'c' or t[0] == t[1] for t, e in x._sorted_items()):
            return self.base_ring().one()
        return self.base_ring().zero()

    class Element(CombinatorialFreeModule.Element):
        """
        An element of a quantum matrix coordinate algebra.
        """
        def bar(self):
            r"""
            Return the image of ``self`` under the bar involution.

            The bar involution is the `\QQ`-algebra anti-automorphism
            defined by `x_{ij} \mapsto x_{ji}` and `q \mapsto q^{-1}`.

            EXAMPLES::

                sage: O = algebras.QuantumMatrixCoordinate(4)
                sage: x = O.an_element()
                sage: x.bar()
                1 + 2*x[1,1] + (q^-16)*x[1,1]^2*x[1,2]^2*x[1,3]^3 + 3*x[1,2]
                sage: x = O.an_element() * O.algebra_generators()[2,4]; x
                x[1,1]^2*x[1,2]^2*x[1,3]^3*x[2,4] + 2*x[1,1]*x[2,4]
                 + 3*x[1,2]*x[2,4] + x[2,4]
                sage: xb = x.bar(); xb
                (q^-16)*x[1,1]^2*x[1,2]^2*x[1,3]^3*x[2,4]
                 + (q^-21-q^-15)*x[1,1]^2*x[1,2]^2*x[1,3]^2*x[1,4]*x[2,3]
                 + (q^-22-q^-18)*x[1,1]^2*x[1,2]*x[1,3]^3*x[1,4]*x[2,2]
                 + (q^-24-q^-20)*x[1,1]*x[1,2]^2*x[1,3]^3*x[1,4]*x[2,1]
                 + 2*x[1,1]*x[2,4] + 3*x[1,2]*x[2,4]
                 + (2*q^-1-2*q)*x[1,4]*x[2,1]
                 + (3*q^-1-3*q)*x[1,4]*x[2,2] + x[2,4]
                sage: xb.bar() == x
                True
            """
            P = self.parent()
            return P.sum(P._bar(c) * P._bar_on_basis(m) for m, c in self)


class QuantumMatrixCoordinateAlgebra(QuantumMatrixCoordinateAlgebra_abstract):
    r"""
    A quantum matrix coordinate algebra.

    Let `R` be a commutative ring. The quantum matrix coordinate algebra
    of `M(m, n)` is the associative algebra over `R[q, q^{-1}]`
    generated by `x_{ij}`, for `i = 1, 2, \ldots, m`, `j = 1, 2, \ldots, n`,
    and subject to the following relations:

    .. MATH::

        \begin{array}{ll}
        x_{it} x_{ij} = q^{-1} x_{ij} x_{it}   & \text{if } j < t, \\
        x_{sj} x_{ij} = q^{-1} x_{ij} x_{sj}   & \text{if } i < s, \\
        x_{st} x_{ij} = x_{ij} x_{st}          & \text{if } i < s, j > t, \\
        x_{st} x_{ij} = x_{ij} x_{st} + (q^{-1} - q) x_{it} x_{sj}
                                               & \text{if } i < s, j < t. \\
        \end{array}

    The quantum matrix coordinate algebra is denoted by
    `\mathcal{O}_q(M(m, n))`. For `m = n`, it is also a bialgebra given by

    .. MATH::

        \Delta(x_{ij}) = \sum_{k=1}^n x_{ik} \otimes x_{kj},
        \varepsilon(x_{ij}) = \delta_{ij}.

    Moreover, there is a central group-like element called the
    *quantum determinant* that is defined by

    .. MATH::

        \det_q = \sum_{\sigma \in S_n} (-q)^{\ell(\sigma)}
        x_{1,\sigma(1)} x_{2,\sigma(2)} \cdots x_{n,\sigma(n)}.

    The quantum matrix coordinate algebra also has natural inclusions
    when restricting to submatrices. That is, let
    `I \subseteq \{1, 2, \ldots, m\}` and `J \subseteq \{1, 2, \ldots, n\}`.
    Then the subalgebra generated by `\{ x_{ij} \mid i \in I, j \in J \}`
    is naturally isomorphic to `\mathcal{O}_q(M(|I|, |J|))`.

    .. NOTE::

        The `q` considered here is `q^2` in some references, e.g., [ZZ2005]_.

    INPUT:

    - ``m`` -- the integer `m`
    - ``n`` -- the integer `n`
    - ``R`` -- (optional) the ring `R` if `q` is not specified
      (the default is `\ZZ`); otherwise the ring containing `q`
    - ``q`` -- (optional) the variable `q`; the default is
      `q \in R[q, q^{-1}]`
    - ``bar`` -- (optional) the involution on the base ring; the
      default is `q \mapsto q^{-1}`

    EXAMPLES:

    We construct `\mathcal{O}_q(M(2,3))` and the variables::

        sage: O = algebras.QuantumMatrixCoordinate(2,3)
        sage: O.inject_variables()
        Defining x11, x12, x13, x21, x22, x23

    We do some basic computations::

        sage: x21 * x11
        (q^-1)*x[1,1]*x[2,1]
        sage: x23 * x12 * x11
        (q^-1)*x[1,1]*x[1,2]*x[2,3] + (q^-2-1)*x[1,1]*x[1,3]*x[2,2]
         + (q^-3-q^-1)*x[1,2]*x[1,3]*x[2,1]

    We construct the maximal quantum minors::

        sage: q = O.q()
        sage: qm12 = x11*x22 - q*x12*x21
        sage: qm13 = x11*x23 - q*x13*x21
        sage: qm23 = x12*x23 - q*x13*x22

    However, unlike for the quantum determinant, they are not central::

        sage: all(qm12 * g == g * qm12 for g in O.algebra_generators())
        False
        sage: all(qm13 * g == g * qm13 for g in O.algebra_generators())
        False
        sage: all(qm23 * g == g * qm23 for g in O.algebra_generators())
        False

    REFERENCES:

    - [FRT1990]_
    - [ZZ2005]_
    """
    @staticmethod
    def __classcall_private__(cls, m, n=None, q=None, bar=None, R=None):
        r"""
        Normalize input to ensure a unique representation.

        EXAMPLES::

            sage: R.<q> = LaurentPolynomialRing(ZZ)
            sage: O1 = algebras.QuantumMatrixCoordinate(4)
            sage: O2 = algebras.QuantumMatrixCoordinate(4, 4, q=q)
            sage: O3 = algebras.QuantumMatrixCoordinate(4, R=ZZ)
            sage: O4 = algebras.QuantumMatrixCoordinate(4, R=R, q=q)
            sage: O1 is O2 and O2 is O3 and O3 is O4
            True
            sage: O5 = algebras.QuantumMatrixCoordinate(4, R=QQ)
            sage: O1 is O5
            False
        """
        if n is None:
            n = m
        return super().__classcall__(cls, m=m, n=n,
                                     q=q, bar=bar,
                                     R=R)

    def __init__(self, m, n, q, bar, R):
        """
        Initialize ``self``.

        TESTS::

            sage: O = algebras.QuantumMatrixCoordinate(4)
            sage: TestSuite(O).run()

            sage: O = algebras.QuantumMatrixCoordinate(10)
            sage: O.variable_names()
            ('x0101', ..., 'x1010')
            sage: O = algebras.QuantumMatrixCoordinate(11,3)
            sage: O.variable_names()
            ('x011', ..., 'x113')
            sage: O = algebras.QuantumMatrixCoordinate(3,11)
            sage: O.variable_names()
            ('x101', ..., 'x311')
        """
        gp_indices = [(i, j) for i in range(1, m + 1) for j in range(1, n + 1)]

        if m == n:
            cat = Bialgebras(R.category()).WithBasis()
        else:
            cat = Algebras(R.category()).WithBasis()

        self._m = m
        QuantumMatrixCoordinateAlgebra_abstract.__init__(self, gp_indices, n, q, bar, R, cat)
        # Set the names
        mb = len(str(m))
        nb = len(str(n))
        base = 'x{{:0>{}}}{{:0>{}}}'.format(mb,nb)
        names = [base.format(*k) for k in gp_indices]
        self._assign_names(names)

    def _repr_(self) -> str:
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: algebras.QuantumMatrixCoordinate(4)
            Quantized coordinate algebra of M(4, 4) with q=q over
             Univariate Laurent Polynomial Ring in q over Integer Ring

            sage: algebras.QuantumMatrixCoordinate(4, 2)
            Quantized coordinate algebra of M(4, 2) with q=q over
             Univariate Laurent Polynomial Ring in q over Integer Ring
        """
        txt = "Quantized coordinate algebra of M({}, {}) with q={} over {}"
        return txt.format(self._m, self._n, self._q, self.base_ring())

    def _latex_(self) -> str:
        r"""
        Return a latex representation of ``self``.

        EXAMPLES::

            sage: O = algebras.QuantumMatrixCoordinate(4)
            sage: latex(O)
            \mathcal{O}_{q}(M(4, 4))
        """
        return "\\mathcal{O}_{%s}(M(%s, %s))" % (self._q, self._m, self._n)

    def m(self):
        """
        Return the value `m`.

        EXAMPLES::

            sage: O = algebras.QuantumMatrixCoordinate(4, 6)
            sage: O.m()
            4
            sage: O = algebras.QuantumMatrixCoordinate(4)
            sage: O.m()
            4
        """
        return self._m

    @cached_method
    def algebra_generators(self) -> Family:
        """
        Return the algebra generators of ``self``.

        EXAMPLES::

            sage: O = algebras.QuantumMatrixCoordinate(2)
            sage: O.algebra_generators()
            Finite family {(1, 1): x[1,1], (1, 2): x[1,2], (2, 1): x[2,1], (2, 2): x[2,2]}
        """
        l = [(i, j) for i in range(1, self._m + 1)
             for j in range(1, self._n + 1)]
        G = self._indices.monoid_generators()
        one = self.base_ring().one()
        return Family(l, lambda x: self.element_class(self, {G[x]: one}))

    def coproduct_on_basis(self, x):
        r"""
        Return the coproduct on the basis element indexed by ``x``.

        EXAMPLES::

            sage: O = algebras.QuantumMatrixCoordinate(4)
            sage: x24 = O.algebra_generators()[2,4]
            sage: O.coproduct_on_basis(x24.leading_support())
            x[2,1] # x[1,4] + x[2,2] # x[2,4] + x[2,3] # x[3,4] + x[2,4] # x[4,4]

        TESTS:

        We check that it is an algebra morphism::

            sage: O = algebras.QuantumMatrixCoordinate(3)
            sage: G = O.algebra_generators()
            sage: all(x.coproduct() * y.coproduct() == (x * y).coproduct()
            ....:     for x in G for y in G)
            True
        """
        if self._m != self._n:
            raise ValueError("undefined for non-square quantum matrices")
        T = self.tensor_square()
        I = self._indices.monoid_generators()
        return T.prod(T.sum_of_monomials((I[t[0], k], I[k, t[1]])
                                         for k in range(1, self._n + 1)) ** e
                      for t, e in x._sorted_items())


class QuantumGL(QuantumMatrixCoordinateAlgebra_abstract):
    r"""
    Quantum coordinate algebra of `GL(n)`.

    The quantum coordinate algebra of `GL(n)`, or quantum `GL(n)`
    for short and denoted by `\mathcal{O}_q(GL(n))`, is the quantum
    coordinate algebra of `M_R(n, n)` with the addition of the
    additional central group-like element `c` which satisfies
    `c d = d c = 1`, where `d` is the quantum determinant.

    Quantum `GL(n)` is a Hopf algebra where `\varepsilon(c) = 1`
    and the antipode `S` is given by the (quantum) matrix inverse.
    That is to say, we have `S(c) = c^-1 = d` and

    .. MATH::

        S(x_{ij}) = c * (-q)^{i-j} * \tilde{t}_{ji},

    where we have the quantum minor

    .. MATH::

        \tilde{t}_{ij} = \sum_{\sigma} (-q)^{\ell(\sigma)}
        x_{1, \sigma(1)} \cdots x_{i-1, \sigma(i-1)} x_{i+1, \sigma(i+1)}
        \cdots x_{n, \sigma(n)}

    with the sum over permutations `\sigma \colon \{1, \ldots, i-1, i+1,
    \ldots n\} \to \{1, \ldots, j-1, j+1, \ldots, n\}`.

    .. SEEALSO::

        :class:`QuantumMatrixCoordinateAlgebra`

    INPUT:

    - ``n`` -- the integer `n`
    - ``R`` -- (optional) the ring `R` if `q` is not specified
      (the default is `\ZZ`); otherwise the ring containing `q`
    - ``q`` -- (optional) the variable `q`; the default is
      `q \in R[q, q^{-1}]`
    - ``bar`` -- (optional) the involution on the base ring; the
      default is `q \mapsto q^{-1}`

    EXAMPLES:

    We construct `\mathcal{O}_q(GL(3))` and the variables::

        sage: O = algebras.QuantumGL(3)
        sage: O.inject_variables()
        Defining x11, x12, x13, x21, x22, x23, x31, x32, x33, c

    We do some basic computations::

        sage: x33 * x12
        x[1,2]*x[3,3] + (q^-1-q)*x[1,3]*x[3,2]
        sage: x23 * x12 * x11
        (q^-1)*x[1,1]*x[1,2]*x[2,3] + (q^-2-1)*x[1,1]*x[1,3]*x[2,2]
         + (q^-3-q^-1)*x[1,2]*x[1,3]*x[2,1]
        sage: c * O.quantum_determinant()
        1

    We verify the quantum determinant is in the center and is group-like::

        sage: qdet = O.quantum_determinant()
        sage: all(qdet * g == g * qdet for g in O.algebra_generators())
        True
        sage: qdet.coproduct() == tensor([qdet, qdet])
        True

    We check that the inverse of the quantum determinant is also in
    the center and group-like::

        sage: all(c * g == g * c for g in O.algebra_generators())
        True
        sage: c.coproduct() == tensor([c, c])
        True

    Moreover, the antipode interchanges the quantum determinant and
    its inverse::

        sage: c.antipode() == qdet
        True
        sage: qdet.antipode() == c
        True

    REFERENCES:

    - [DD1991]_
    - [Kar1993]_
    """
    @staticmethod
    def __classcall_private__(cls, n, q=None, bar=None, R=None):
        """
        Normalize input to ensure a unique representation.

        EXAMPLES::

            sage: R.<q> = LaurentPolynomialRing(ZZ)
            sage: O1 = algebras.QuantumGL(4)
            sage: O2 = algebras.QuantumGL(4, R=ZZ)
            sage: O3 = algebras.QuantumGL(4, R=R, q=q)
            sage: O1 is O2 and O2 is O3
            True
            sage: O4 = algebras.QuantumGL(4, R=QQ)
            sage: O1 is O4
            False
        """
        return super().__classcall__(cls, n=n, q=q, bar=bar, R=R)

    def __init__(self, n, q, bar, R):
        """
        Initialize ``self``.

        TESTS::

            sage: O = algebras.QuantumGL(2)
            sage: elts = list(O.algebra_generators())
            sage: elts += [O.quantum_determinant(), O.an_element()]
            sage: TestSuite(O).run(elements=elts) # long time
        """
        # Set the names
        gp_indices = [(i, j) for i in range(1, n + 1) for j in range(1, n + 1)]
        gp_indices.append('c')
        cat = HopfAlgebras(R.category()).WithBasis()
        QuantumMatrixCoordinateAlgebra_abstract.__init__(self, gp_indices, n, q,
                                                         bar, R, cat,
                                                         indices_key=_generator_key)
        names = ['x{}{}'.format(*k) for k in gp_indices[:-1]]
        names.append('c')
        self._assign_names(names)

    def _repr_(self) -> str:
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: algebras.QuantumGL(4)
            Quantized coordinate algebra of GL(4) with q=q over
             Univariate Laurent Polynomial Ring in q over Integer Ring
        """
        txt = "Quantized coordinate algebra of GL({}) with q={} over {}"
        return txt.format(self._n, self._q, self.base_ring())

    def _latex_(self) -> str:
        r"""
        Return a latex representation of ``self``.

        EXAMPLES::

            sage: O = algebras.QuantumGL(4)
            sage: latex(O)
            \mathcal{O}_{q}(GL(4))
        """
        return "\\mathcal{O}_{%s}(GL(%s))" % (self._q, self._n)

    @cached_method
    def algebra_generators(self):
        """
        Return the algebra generators of ``self``.

        EXAMPLES::

            sage: O = algebras.QuantumGL(2)
            sage: O.algebra_generators()
            Finite family {(1, 1): x[1,1], (1, 2): x[1,2], (2, 1): x[2,1], (2, 2): x[2,2], 'c': c}
        """
        l = [(i, j) for i in range(1, self._n + 1)
             for j in range(1, self._n + 1)]
        l.append('c')
        G = self._indices.monoid_generators()
        one = self.base_ring().one()
        return Family(l, lambda x: self.element_class(self, {G[x]: one}))

    @lazy_attribute
    def _qdet_cancel_monomial(self):
        """
        Return the trailing monomial of the quantum determinant.

        EXAMPLES::

            sage: O = algebras.QuantumGL(2)
            sage: O._qdet_cancel_monomial
            F[(1, 1)]*F[(2, 2)]
        """
        I = self._indices
        gens = I.monoid_generators()
        return I.prod(gens[i, i] for i in range(1, self._n + 1))

    @lazy_attribute
    def _qdet_remaining(self):
        r"""
        Return the remaining terms when cancelling the leading term.

        Consider `d = m + L`, where `m` is the leading term of the
        quantum determinant `d`. Then we have `c d = cm + cL = 1`,
        which we rewrite as `cm = 1 - cL`. This lazy attribute
        is `1 - cL`.

        EXAMPLES::

            sage: O = algebras.QuantumGL(2)
            sage: O._qdet_remaining
            1 + q*c*x[1,2]*x[2,1]
        """
        temp = self.monomial(self._qdet_cancel_monomial) - self.quantum_determinant()
        c = self._indices.monoid_generators()['c']
        ret = {c * mon: coeff for mon, coeff in temp}
        return self._from_dict(ret, remove_zeros=False) + self.one()

    def product_on_basis(self, a, b):
        r"""
        Return the product of basis elements indexed by ``a`` and ``b``.

        EXAMPLES::

            sage: O = algebras.QuantumGL(2)
            sage: I = O.indices().monoid_generators()
            sage: O.product_on_basis(I[1,1], I[2,2])
            x[1,1]*x[2,2]
            sage: O.product_on_basis(I[2,2], I[1,1])
            x[1,1]*x[2,2] + (q^-1-q)*x[1,2]*x[2,1]

        TESTS::

            sage: x11,x12,x21,x22,c = O.algebra_generators()
            sage: x11 * x22
            x[1,1]*x[2,2]
            sage: x22 * x12
            (q^-1)*x[1,2]*x[2,2]
            sage: x22 * x11
            x[1,1]*x[2,2] + (q^-1-q)*x[1,2]*x[2,1]
            sage: c * (x11 * O.quantum_determinant())
            x[1,1]
        """
        I = self._indices
        c_exp = 0
        if 'c' in a._monomial:
            da = dict(a._monomial)  # Make a copy
            c_exp += da.pop('c')
            a = I(da)
        if 'c' in b._monomial:
            db = dict(b._monomial)  # Make a copy
            c_exp += db.pop('c')
            b = I(db)
        # a and b contain no powers of c
        p = super().product_on_basis(a, b)
        if c_exp == 0:
            return p
        c = self._indices.monoid_generators()['c']
        ret = {}
        other = self.zero()
        for mon, coeff in p:
            try:
                # Given that cz = R and we have a monomial ab, we need to
                #   rewrite zx in terms of ab plus lower order terms L:
                # zx = X * ab + L
                # c * zx = R * x = c * X * ab + c * L
                # c * ab = (R * x - c * L) / X
                rem = self.monomial(mon // self._qdet_cancel_monomial)
                L = self.monomial(self._qdet_cancel_monomial) * rem
                co = L[mon]
                del L._monomial_coefficients[mon]
                temp = self.term(c ** (c_exp - 1), coeff) * self._qdet_remaining * rem
                if L != self.zero():
                    temp -= self.term(c ** c_exp, coeff) * L
                for k in temp._monomial_coefficients:
                    temp._monomial_coefficients[k] //= co
                other += temp
            except ValueError:  # We cannot cancel, so we just add on the correct power of c
                ret[c ** c_exp * mon] = coeff
        return self._from_dict(ret, remove_zeros=False) + other

    @cached_method
    def _antipode_on_generator(self, i, j):
        """
        Return the antipode on the generator indexed by ``(i, j)``.

        EXAMPLES::

            sage: O = algebras.QuantumGL(2)
            sage: [[O._antipode_on_generator(i, j) for i in [1,2]] for j in [1,2]]
            [[c*x[2,2], -q*c*x[2,1]],
             [-(q^-1)*c*x[1,2], c*x[1,1]]]
        """
        from sage.combinat.permutation import Permutations
        q = self._q
        I = list(range(1, j)) + list(range(j + 1, self._n + 1))

        def lift(p):
            return [val if val < i else val + 1 for val in p]
        gens = self.algebra_generators()
        t_tilde = self.sum((-q) ** p.length() * gens['c'] *
                           self.prod(gens[I[k], val]
                                     for k, val in enumerate(lift(p)))
                           for p in Permutations(self._n - 1))
        return (-q) ** (i - j) * t_tilde

    def antipode_on_basis(self, x):
        r"""
        Return the antipode of the basis element indexed by ``x``.

        EXAMPLES::

            sage: O = algebras.QuantumGL(3)
            sage: x = O.indices().monoid_generators()
            sage: O.antipode_on_basis(x[1,2])
            -(q^-1)*c*x[1,2]*x[3,3] + c*x[1,3]*x[3,2]
            sage: O.antipode_on_basis(x[2,2])
            c*x[1,1]*x[3,3] - q*c*x[1,3]*x[3,1]
            sage: O.antipode_on_basis(x['c']) == O.quantum_determinant()
            True
        """
        ret = self.one()
        for k, e in reversed(x._sorted_items()):
            if k == 'c':
                ret *= self.quantum_determinant() ** e
            else:
                ret *= self._antipode_on_generator(*k) ** e
        return ret

    def coproduct_on_basis(self, x):
        r"""
        Return the coproduct on the basis element indexed by ``x``.

        EXAMPLES::

            sage: O = algebras.QuantumGL(3)
            sage: x = O.indices().monoid_generators()
            sage: O.coproduct_on_basis(x[1,2])
            x[1,1] # x[1,2] + x[1,2] # x[2,2] + x[1,3] # x[3,2]
            sage: O.coproduct_on_basis(x[2,2])
            x[2,1] # x[1,2] + x[2,2] # x[2,2] + x[2,3] # x[3,2]
            sage: O.coproduct_on_basis(x['c'])
            c # c
        """
        T = self.tensor_square()
        I = self._indices.monoid_generators()
        return T.prod(T.sum_of_monomials((I[t[0], k], I[k, t[1]])
                                         for k in range(1, self._n + 1)) ** e
                      if t != 'c' else T.monomial((I['c'], I['c'])) ** e
                      for t, e in x._sorted_items())


def _generator_key(t):
    """
    Helper function to make ``'c'`` less that all other indices for
    sorting the monomials in :class:`QuantumGL`.

    INPUT:

    - ``t`` -- tuple (index, exponent)

    OUTPUT: a tuple made from the index only

    EXAMPLES::

        sage: from sage.algebras.quantum_matrix_coordinate_algebra import _generator_key as k
        sage: k(((1,2),1)) < k(('c',1))
        False
        sage: k(((1,2),1)) < k(((1,3),1))
        True
        sage: k(((1,2),1)) < k(((3,1),1))
        True
        sage: k(('c',2)) < k(((1,1),1))
        True
    """
    t = t[0]
    if isinstance(t, tuple):
        return t
    return ()



# distutils: language = c++
# distutils: libraries = ec NTL_LIBRARIES pari gmp m
# distutils: extra_compile_args = NTL_CFLAGS
# distutils: include_dirs = NTL_INCDIR
# distutils: library_dirs = NTL_LIBDIR
# distutils: extra_link_args = NTL_LIBEXTRA


from libcpp.map cimport map
from libcpp.vector cimport vector
from libcpp.pair cimport pair
from sage.libs.ntl.types cimport ZZ_c


# NOTE: eclib used to have specific dependencies, so that they had to
# be included in a specific order. Although this is no longer the
# case, we start by listing all relevant include files in the correct
# order.

cdef extern from "eclib/vector.h": pass
cdef extern from "eclib/xmod.h": pass
cdef extern from "eclib/svector.h": pass
cdef extern from "eclib/matrix.h": pass
cdef extern from "eclib/smatrix.h": pass
cdef extern from "eclib/interface.h": pass
cdef extern from "eclib/rat.h": pass
cdef extern from "eclib/bigrat.h": pass
cdef extern from "eclib/isogs.h": pass
cdef extern from "eclib/curve.h": pass
cdef extern from "eclib/descent.h": pass
cdef extern from "eclib/egr.h": pass
cdef extern from "eclib/htconst.h": pass
cdef extern from "eclib/moddata.h": pass
cdef extern from "eclib/symb.h": pass
cdef extern from "eclib/homspace.h": pass
cdef extern from "eclib/oldforms.h": pass
cdef extern from "eclib/newforms.h": pass


# Now the actual declarations, where the order no longer matters
cdef extern from "eclib/vector.h":
    ctypedef int scalar

    cdef cppclass vec:
        scalar operator[](long)

cdef extern from "eclib/svector.h":
    cdef cppclass svec:
        vec as_vec()
        scalar elem(int) const
        map[int,scalar].iterator begin()
        map[int,scalar].iterator end()

cdef extern from "eclib/matrix.h":
    cdef cppclass mat:
        mat()
        mat(mat m)
        scalar sub(long, long)
        long nrows()
        long ncols()
        long rank()

    mat addscalar(mat M, scalar)

cdef extern from "eclib/smatrix.h":
    cdef cppclass smat:
        smat()
        smat(smat m)
        scalar sub(long, long)
        long nrows()
        long ncols()
        long rank()
        scalar elem(int, int) const
        svec row(int) const

cdef extern from "eclib/interface.h":
    ctypedef ZZ_c bigint
    int I2int(bigint)

cdef extern from "eclib/rat.h":
    cdef cppclass rational:
        rational()
        rational(long num, long den)
    cdef long rational_num "num"(rational q)
    cdef long rational_den "den"(rational q)

cdef extern from "eclib/bigrat.h":
    cdef cppclass bigrational:
        pass
    cdef bigint bigrational_num "num"(bigrational q)
    cdef bigint bigrational_den "den"(bigrational q)

cdef extern from "eclib/curve.h":
    cdef cppclass Curve:
        Curve()
        Curve(bigint aa1, bigint aa2, bigint aa3, bigint aa4, bigint aa6)
        void getai(bigint a1, bigint a2, bigint a3, bigint a4, bigint a6)

    cdef cppclass Curvedata:
        Curvedata()
        Curvedata(Curve C, int m)
        Curvedata(bigint a1, bigint a2, bigint a3, bigint a4, bigint a6,
                int min_on_init)
        void getai(bigint a1, bigint a2, bigint a3, bigint a4, bigint a6)

    cdef cppclass CurveRed:
        CurveRed()
        CurveRed(Curvedata CD)

    bigint getconductor(CurveRed CR)

cdef extern from "eclib/descent.h":
    cdef cppclass two_descent:
        two_descent(Curvedata* curve,
                    int verb, int sel,
                    long firstlim, long secondlim,
                    long n_aux, int second_descent)

    cdef cppclass mw:
        mw(Curvedata* curve, int verb, int pp, int maxr)

cdef extern from "eclib/homspace.h":
    cdef cppclass homspace:
        long modulus
        int plusflag # sign
        int cuspidal

        homspace(long n, int hp, int hcusp, int verbose)

        int h1dim()
        int h1cuspdim()
        mat heckeop(long p, int dual, int display)
        vec heckeop_col(long p, int j, int display)
        smat s_heckeop(long p, int dual, int display)
        long h1ncusps()

cdef extern from "eclib/newforms.h":
    cdef cppclass newforms:
        vector[newform] nflist
        int verbose
        int basisflag
        long maxdepth
        long cuspidal
        long sign

        newforms(long n, int disp)

        void createfromcurve(int sign, CurveRed CR, int nap)
        void display()
        # Here i is the index of the relevant newform in the space,
        # which for us will always be 0:
        rational plus_modular_symbol(rational r, int i, int base_at_infinity)
        rational minus_modular_symbol(rational r, int i, int base_at_infinity)
        pair[rational,rational] full_modular_symbol(rational r, int i, int base_at_infinity)

    cdef cppclass newform:
        newforms* nf



# distutils: language = c++
r"""
Pairing Heap

This module proposes several implementations of the pairing heap data structure
[FSST1986]_. See the :wikipedia:`Pairing_heap` for more information on this
min-heap data structure.

- :class:`PairingHeap_of_n_integers`: a pairing heap data structure with fixed
  capacity `n`. Its items are integers in the range `[0, n-1]`. Values can be of
  any type equipped with a comparison method (``<=``).

- :class:`PairingHeap_of_n_hashables`: a pairing heap data structure with fixed
  capacity `n`. Its items can be of any hashable type. Values can be of any type
  equipped with a comparison method (``<=``).

- ``PairingHeap``: interface to a pairing heap data structure written in C++.
  The advantages of this data structure are that: its capacity is unbounded;
  items can be of any hashable type equipped with a hashing method that can be
  supported by ``std::unordered_map``; values can be of any specified type
  equipped with a comparison method (``<=``). This data structure is for
  internal use and therefore cannot be accessed from a shell.

EXAMPLES:

Pairing heap of `n` integers in the range `[0, n-1]`::

    sage: from sage.data_structures.pairing_heap import PairingHeap_of_n_integers
    sage: P = PairingHeap_of_n_integers(10); P
    PairingHeap_of_n_integers: capacity 10, size 0
    sage: P.push(1, 3)
    sage: P.push(2, 2)
    sage: P
    PairingHeap_of_n_integers: capacity 10, size 2
    sage: P.top()
    (2, 2)
    sage: P.decrease(1, 1)
    sage: P.top()
    (1, 1)
    sage: P.pop()
    sage: P.top()
    (2, 2)

    sage: P = PairingHeap_of_n_integers(10)
    sage: P.push(1, (2, 'a'))
    sage: P.push(2, (2, 'b'))
    sage: P.top()
    (1, (2, 'a'))

Pairing heap of `n` hashables::

    sage: from sage.data_structures.pairing_heap import PairingHeap_of_n_hashables
    sage: P = PairingHeap_of_n_hashables(10); P
    PairingHeap_of_n_hashables: capacity 10, size 0
    sage: P.push(1, 3)
    sage: P.push('b', 2)
    sage: P.push((1, 'abc'), 4)
    sage: P.top()
    ('b', 2)
    sage: P.decrease((1, 'abc'), 1)
    sage: P.top()
    ((1, 'abc'), 1)
    sage: P.pop()
    sage: P.top()
    ('b', 2)

    sage: # needs sage.graphs
    sage: P = PairingHeap_of_n_hashables(10)
    sage: P.push(('a', 1), (2, 'b'))
    sage: P.push(2, (2, 'a'))
    sage: g = Graph(2, immutable=True)
    sage: P.push(g, (3, 'z'))
    sage: P.top()
    (2, (2, 'a'))
    sage: P.decrease(g, (1, 'z'))
    sage: P.top()
    (Graph on 2 vertices, (1, 'z'))
    sage: while P:
    ....:     print(P.top())
    ....:     P.pop()
    (Graph on 2 vertices, (1, 'z'))
    (2, (2, 'a'))
    (('a', 1), (2, 'b'))
"""
# ******************************************************************************
# Copyright (C) 2024 David Coudert <david.coudert@inria.fr>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ******************************************************************************


from libcpp.pair cimport pair
from cpython.ref cimport PyObject, Py_INCREF, Py_XDECREF
from cysignals.signals cimport sig_on, sig_off, sig_check
from cysignals.memory cimport check_allocarray, sig_free
from sage.data_structures.bitset_base cimport (bitset_init, bitset_free,
                                               bitset_add, bitset_remove,
                                               bitset_in)
from sage.misc.prandom import shuffle


# ==============================================================================
# Class PairingHeap_class
# ==============================================================================

cdef class PairingHeap_class:
    r"""
    Common class and methods for :class:`PairingHeap_of_n_integers` and
    :class:`PairingHeap_of_n_hashables`.
    """

    def __repr__(self):
        r"""
        Return a string representing ``self``.

        EXAMPLES::

            sage: from sage.data_structures.pairing_heap import PairingHeap_of_n_integers
            sage: P = PairingHeap_of_n_integers(5); P
            PairingHeap_of_n_integers: capacity 5, size 0
            sage: P.push(1, 2)
            sage: P
            PairingHeap_of_n_integers: capacity 5, size 1
        """
        return f"{type(self).__name__}: capacity {self.n}, size {len(self)}"

    def __bool__(self):
        r"""
        Check whether ``self`` is not empty.

        EXAMPLES::

            sage: from sage.data_structures.pairing_heap import PairingHeap_of_n_integers
            sage: P = PairingHeap_of_n_integers(5)
            sage: 'not empty' if P else 'empty'
            'empty'
            sage: P.push(1, 2)
            sage: 'not empty' if P else 'empty'
            'not empty'
        """
        return self.root != NULL

    cpdef bint empty(self) noexcept:
        r"""
        Check whether the heap is empty.

        EXAMPLES::

            sage: from sage.data_structures.pairing_heap import PairingHeap_of_n_integers
            sage: P = PairingHeap_of_n_integers(5)
            sage: P.empty()
            True
            sage: P.push(1, 2)
            sage: P.empty()
            False
        """
        return self.root == NULL

    cpdef bint full(self) noexcept:
        r"""
        Check whether the heap is full.

        EXAMPLES::

            sage: from sage.data_structures.pairing_heap import PairingHeap_of_n_integers
            sage: P = PairingHeap_of_n_integers(2)
            sage: P.full()
            False
            sage: P.push(0, 2)
            sage: P.push(1, 3)
            sage: P.full()
            True
        """
        return self.n == self.number_of_items

    cpdef size_t capacity(self) noexcept:
        r"""
        Return the maximum capacity of the heap.

        EXAMPLES::

            sage: from sage.data_structures.pairing_heap import PairingHeap_of_n_integers
            sage: P = PairingHeap_of_n_integers(5)
            sage: P.capacity()
            5
            sage: P.push(1, 2)
            sage: P.capacity()
            5
        """
        return self.n

    cpdef size_t size(self) noexcept:
        r"""
        Return the number of items in the heap.

        EXAMPLES::

            sage: from sage.data_structures.pairing_heap import PairingHeap_of_n_integers
            sage: P = PairingHeap_of_n_integers(5)
            sage: P.size()
            0
            sage: P.push(1, 2)
            sage: P.size()
            1

        One may also use Python's ``__len__``::

            sage: len(P)
            1
        """
        return self.number_of_items

    def __len__(self):
        r"""
        Return the number of items in the heap.

        EXAMPLES::

            sage: from sage.data_structures.pairing_heap import PairingHeap_of_n_integers
            sage: P = PairingHeap_of_n_integers(5)
            sage: len(P)
            0
            sage: P.push(1, 2)
            sage: len(P)
            1
        """
        return self.number_of_items

    cpdef tuple top(self):
        r"""
        Return the top pair (item, value) of the heap.

        EXAMPLES::

            sage: from sage.data_structures.pairing_heap import PairingHeap_of_n_integers
            sage: P = PairingHeap_of_n_integers(5)
            sage: P.push(1, 2)
            sage: P.top()
            (1, 2)
            sage: P.push(3, 1)
            sage: P.top()
            (3, 1)

            sage: P = PairingHeap_of_n_integers(3)
            sage: P.top()
            Traceback (most recent call last):
            ...
            ValueError: trying to access the top of an empty heap
        """
        raise NotImplementedError()

    cpdef object top_value(self):
        r"""
        Return the value of the top item of the heap.

        EXAMPLES::

            sage: from sage.data_structures.pairing_heap import PairingHeap_of_n_integers
            sage: P = PairingHeap_of_n_integers(5)
            sage: P.push(1, 2)
            sage: P.top()
            (1, 2)
            sage: P.top_value()
            2

            sage: P = PairingHeap_of_n_integers(3)
            sage: P.top_value()
            Traceback (most recent call last):
            ...
            ValueError: trying to access the top of an empty heap
        """
        if not self:
            raise ValueError("trying to access the top of an empty heap")
        return <object>self.root.value

    cpdef void pop(self) noexcept:
        r"""
        Remove the top item from the heap.

        If the heap is already empty, we do nothing.

        EXAMPLES::

            sage: from sage.data_structures.pairing_heap import PairingHeap_of_n_integers
            sage: P = PairingHeap_of_n_integers(5); P
            PairingHeap_of_n_integers: capacity 5, size 0
            sage: P.push(1, 2); P
            PairingHeap_of_n_integers: capacity 5, size 1
            sage: P.pop(); P
            PairingHeap_of_n_integers: capacity 5, size 0
            sage: P.pop(); P
            PairingHeap_of_n_integers: capacity 5, size 0
        """
        raise NotImplementedError()


# ==============================================================================
# Class PairingHeap_of_n_integers
# ==============================================================================

cdef class PairingHeap_of_n_integers(PairingHeap_class):
    r"""
    Pairing Heap for items in range `[0, n - 1]`.

    EXAMPLES::

        sage: from sage.data_structures.pairing_heap import PairingHeap_of_n_integers
        sage: P = PairingHeap_of_n_integers(5); P
        PairingHeap_of_n_integers: capacity 5, size 0
        sage: P.push(1, 3)
        sage: P.push(2, 2)
        sage: P
        PairingHeap_of_n_integers: capacity 5, size 2
        sage: P.top()
        (2, 2)
        sage: P.decrease(1, 1)
        sage: P.top()
        (1, 1)
        sage: P.pop()
        sage: P.top()
        (2, 2)

    TESTS::

        sage: from sage.data_structures.pairing_heap import PairingHeap_of_n_integers
        sage: P = PairingHeap_of_n_integers(0)
        Traceback (most recent call last):
        ...
        ValueError: the capacity of the heap must be strictly positive
        sage: P = PairingHeap_of_n_integers(1); P
        PairingHeap_of_n_integers: capacity 1, size 0
        sage: P = PairingHeap_of_n_integers(5)
        sage: P.push(11, 3)
        Traceback (most recent call last):
        ...
        ValueError: item must be in range [0, 4]
    """

    def __init__(self, size_t n):
        r"""
        Construct the ``PairingHeap_of_n_integers`` where items are integers
        from `0` to `n-1`.

        INPUT:

        - ``n`` -- strictly positive integer; the maximum number of items in the heap

        EXAMPLES::

            sage: from sage.data_structures.pairing_heap import PairingHeap_of_n_integers
            sage: P = PairingHeap_of_n_integers(5); P
            PairingHeap_of_n_integers: capacity 5, size 0
            sage: P.push(1, 2); P
            PairingHeap_of_n_integers: capacity 5, size 1
            sage: P.push(2, 3); P
            PairingHeap_of_n_integers: capacity 5, size 2
            sage: P.pop(); P
            PairingHeap_of_n_integers: capacity 5, size 1
            sage: P.push(10, 1)
            Traceback (most recent call last):
            ...
            ValueError: item must be in range [0, 4]
            sage: PairingHeap_of_n_integers(1)
            PairingHeap_of_n_integers: capacity 1, size 0

        TESTS::

            sage: PairingHeap_of_n_integers(0)
            Traceback (most recent call last):
            ...
            ValueError: the capacity of the heap must be strictly positive
            sage: P = PairingHeap_of_n_integers(10)
            sage: P.push(1, 1); P.push(7, 0); P.push(0, 4); P.pop(); P.push(5, 5)
            sage: TestSuite(P).run(skip="_test_pickling")
        """
        if not n:
            raise ValueError("the capacity of the heap must be strictly positive")
        self.n = n
        self.root = NULL
        self.nodes = <PairingHeapNodePy *>check_allocarray(n, sizeof(PairingHeapNodePy))
        bitset_init(self.active, n)
        self.number_of_items = 0

    def __dealloc__(self):
        """
        Deallocate ``self``.

        EXAMPLES::

            sage: from sage.data_structures.pairing_heap import PairingHeap_of_n_integers
            sage: P = PairingHeap_of_n_integers(5)
            sage: del P
        """
        sig_free(self.nodes)
        bitset_free(self.active)

    cpdef void push(self, size_t item, object value) except *:
        r"""
        Insert an item into the heap with specified value (priority).

        INPUT:

        - ``item`` -- nonnegative integer; the item to consider

        - ``value`` -- the value associated with ``item``

        EXAMPLES::

            sage: from sage.data_structures.pairing_heap import PairingHeap_of_n_integers
            sage: P = PairingHeap_of_n_integers(5)
            sage: P.push(1, 2)
            sage: P.top()
            (1, 2)
            sage: P.push(3, 1)
            sage: P.top()
            (3, 1)

        TESTS::

            sage: from sage.data_structures.pairing_heap import PairingHeap_of_n_integers
            sage: P = PairingHeap_of_n_integers(5)
            sage: P.push(1, 2)
            sage: P.push(1, 2)
            Traceback (most recent call last):
            ...
            ValueError: 1 is already in the heap
            sage: P.push(11, 2)
            Traceback (most recent call last):
            ...
            ValueError: item must be in range [0, 4]
            sage: P.push(-1, 0)
            Traceback (most recent call last):
            ...
            OverflowError: can't convert negative value to size_t
        """
        if item >= self.n:
            raise ValueError(f"item must be in range [0, {self.n - 1}]")
        if item in self:
            raise ValueError(f"{item} is already in the heap")

        cdef PairingHeapNodePy * p = self.nodes + item
        Py_INCREF(value)
        p.value = <PyObject *>value
        p.prev = p.next = p.child = NULL
        if self.root == NULL:
            self.root = p
        else:
            self.root = PairingHeapNodePy._merge(self.root, p)
        bitset_add(self.active, item)
        self.number_of_items += 1

    cpdef tuple top(self):
        r"""
        Return the top pair (item, value) of the heap.

        EXAMPLES::

            sage: from sage.data_structures.pairing_heap import PairingHeap_of_n_integers
            sage: P = PairingHeap_of_n_integers(5)
            sage: P.push(1, 2)
            sage: P.top()
            (1, 2)
            sage: P.push(3, 1)
            sage: P.top()
            (3, 1)

            sage: P = PairingHeap_of_n_integers(3)
            sage: P.top()
            Traceback (most recent call last):
            ...
            ValueError: trying to access the top of an empty heap
        """
        if not self:
            raise ValueError("trying to access the top of an empty heap")
        return self.root - self.nodes, <object>self.root.value

    cpdef size_t top_item(self) except *:
        r"""
        Return the top item of the heap.

        EXAMPLES::

            sage: from sage.data_structures.pairing_heap import PairingHeap_of_n_integers
            sage: P = PairingHeap_of_n_integers(5)
            sage: P.push(1, 2)
            sage: P.top()
            (1, 2)
            sage: P.top_item()
            1

            sage: P = PairingHeap_of_n_integers(3)
            sage: P.top_item()
            Traceback (most recent call last):
            ...
            ValueError: trying to access the top of an empty heap
        """
        if not self:
            raise ValueError("trying to access the top of an empty heap")
        return self.root - self.nodes

    cpdef void pop(self) noexcept:
        r"""
        Remove the top item from the heap.

        If the heap is already empty, we do nothing.

        EXAMPLES::

            sage: from sage.data_structures.pairing_heap import PairingHeap_of_n_integers
            sage: P = PairingHeap_of_n_integers(5); P
            PairingHeap_of_n_integers: capacity 5, size 0
            sage: P.push(1, 2); P
            PairingHeap_of_n_integers: capacity 5, size 1
            sage: P.push(2, 3); P
            PairingHeap_of_n_integers: capacity 5, size 2
            sage: P.pop(); P
            PairingHeap_of_n_integers: capacity 5, size 1
            sage: P.pop(); P
            PairingHeap_of_n_integers: capacity 5, size 0
            sage: P.pop(); P
            PairingHeap_of_n_integers: capacity 5, size 0
        """
        if not self:
            return
        cdef size_t item = self.top_item()
        Py_XDECREF(<PyObject *>self.nodes[item].value)
        bitset_remove(self.active, item)
        self.number_of_items -= 1
        self.root = PairingHeapNodePy._pair(self.root.child)

    cpdef void decrease(self, size_t item, object new_value) except *:
        r"""
        Decrease the value of specified item.

        This method is more permissive than it should as it can also be used to
        push an item in the heap.

        INPUT:

        - ``item`` -- nonnegative integer; the item to consider

        - ``new_value`` -- the new value for ``item``

        EXAMPLES::

            sage: from sage.data_structures.pairing_heap import PairingHeap_of_n_integers
            sage: P = PairingHeap_of_n_integers(5)
            sage: 3 in P
            False
            sage: P.decrease(3, 33)
            sage: 3 in P
            True
            sage: P.top()
            (3, 33)
            sage: P.push(1, 10)
            sage: P.top()
            (1, 10)
            sage: P.decrease(3, 7)
            sage: P.top()
            (3, 7)

        TESTS::

            sage: P = PairingHeap_of_n_integers(5)
            sage: P.push(1, 3)
            sage: P.decrease(1, 2)
            sage: P.decrease(1, 2)
            Traceback (most recent call last):
            ...
            ValueError: the new value must be less than the current value
        """
        if item >= self.n:
            raise ValueError(f"item must be in range [0, {self.n - 1}]")
        cdef PairingHeapNodePy * p
        if bitset_in(self.active, item):
            p = self.nodes + item
            if <object>p.value <= new_value:
                raise ValueError("the new value must be less than the current value")
            Py_XDECREF(<PyObject *>p.value)
            Py_INCREF(new_value)
            p.value = <PyObject *>new_value
            if p.prev != NULL:
                PairingHeapNodePy._unlink(p)
                self.root = PairingHeapNodePy._merge(self.root, p)
        else:
            self.push(item, new_value)

    def __contains__(self, size_t item):
        r"""
        Check whether the specified item is in the heap.

        INPUT:

        - ``item`` -- nonnegative integer; the item to consider

        EXAMPLES::

            sage: from sage.data_structures.pairing_heap import PairingHeap_of_n_integers
            sage: P = PairingHeap_of_n_integers(5)
            sage: 3 in P
            False
            sage: P.push(3, 33)
            sage: 3 in P
            True
            sage: 100 in P
            False
        """
        if item >= self.n:
            return False
        return bitset_in(self.active, item)

    contains = __contains__

    cpdef object value(self, size_t item):
        r"""
        Return the value associated with the item.

        INPUT:

        - ``item`` -- nonnegative integer; the item to consider

        EXAMPLES::

            sage: from sage.data_structures.pairing_heap import PairingHeap_of_n_integers
            sage: P = PairingHeap_of_n_integers(5)
            sage: P.push(3, 33)
            sage: P.push(1, 10)
            sage: P.value(3)
            33
            sage: P.value(7)
            Traceback (most recent call last):
            ...
            ValueError: 7 is not in the heap
        """
        if item not in self:
            raise ValueError(f"{item} is not in the heap")
        return <object>self.nodes[item].value


# ==============================================================================
# Class PairingHeap_of_n_hashables
# ==============================================================================

cdef class PairingHeap_of_n_hashables(PairingHeap_class):
    r"""
    Pairing Heap for `n` hashable items.

    EXAMPLES::

        sage: from sage.data_structures.pairing_heap import PairingHeap_of_n_hashables
        sage: P = PairingHeap_of_n_hashables(5); P
        PairingHeap_of_n_hashables: capacity 5, size 0
        sage: P.push(1, 3)
        sage: P.push('abc', 2); P
        PairingHeap_of_n_hashables: capacity 5, size 2
        sage: P.top()
        ('abc', 2)
        sage: P.decrease(1, 1)
        sage: P.top()
        (1, 1)
        sage: P.pop()
        sage: P.top()
        ('abc', 2)

        sage: P = PairingHeap_of_n_hashables(5)
        sage: P.push(1, (2, 3))
        sage: P.push('a', (2, 2))
        sage: P.push('b', (3, 3))
        sage: P.push('c', (2, 1))
        sage: P.top()
        ('c', (2, 1))
        sage: P.push(Graph(2, immutable=True), (1, 7))
        sage: P.top()
        (Graph on 2 vertices, (1, 7))
        sage: P.decrease('b', (1, 5))
        sage: P.top()
        ('b', (1, 5))

    TESTS::

        sage: from sage.data_structures.pairing_heap import PairingHeap_of_n_hashables
        sage: P = PairingHeap_of_n_hashables(0)
        Traceback (most recent call last):
        ...
        ValueError: the capacity of the heap must be strictly positive
        sage: P = PairingHeap_of_n_hashables(1); P
        PairingHeap_of_n_hashables: capacity 1, size 0
        sage: P.push(11, 3)
        sage: P.push(12, 4)
        Traceback (most recent call last):
        ...
        ValueError: the heap is full

        sage: P = PairingHeap_of_n_hashables(10)
        sage: P.push(1, 'John')
        sage: P.push(4, 42)
        Traceback (most recent call last):
        ...
        TypeError: unsupported operand parent(s) for >=: 'Integer Ring' and '<class 'str'>'
    """

    def __init__(self, size_t n):
        r"""
        Construct the ``PairingHeap_of_n_hashables``.

        This pairing heap has a maximum capacity of `n` items and each item is a
        hashable object.

        INPUT:

        - ``n`` -- strictly positive integer; the maximum number of items in the heap

        EXAMPLES::

            sage: from sage.data_structures.pairing_heap import PairingHeap_of_n_hashables
            sage: P = PairingHeap_of_n_hashables(2); P
            PairingHeap_of_n_hashables: capacity 2, size 0
            sage: P.push(1, 2)
            sage: P
            PairingHeap_of_n_hashables: capacity 2, size 1
            sage: P.push(2, 3)
            sage: P
            PairingHeap_of_n_hashables: capacity 2, size 2
            sage: P.full()
            True
            sage: P.push(10, 1)
            Traceback (most recent call last):
            ...
            ValueError: the heap is full
            sage: P.pop()
            sage: P
            PairingHeap_of_n_hashables: capacity 2, size 1
            sage: P.push(10, 1)

    TESTS::

            sage: P = PairingHeap_of_n_hashables(0)
            Traceback (most recent call last):
            ...
            ValueError: the capacity of the heap must be strictly positive
            sage: P = PairingHeap_of_n_hashables(1); P
            PairingHeap_of_n_hashables: capacity 1, size 0
            sage: P = PairingHeap_of_n_hashables(6)
            sage: P.push(1, -0.5); P.push(frozenset(), 1); P.pop(); P.push('a', 3.5)
            sage: TestSuite(P).run(skip="_test_pickling")
        """
        if not n:
            raise ValueError("the capacity of the heap must be strictly positive")
        self.n = n
        self.root = NULL
        self.nodes = <PairingHeapNodePy *>check_allocarray(n, sizeof(PairingHeapNodePy))
        self.number_of_items = 0
        self._int_to_item = [None] * n
        self._item_to_int = dict()
        self.free_idx = list(range(n))

    def __dealloc__(self):
        """
        Deallocate ``self``.

        EXAMPLES::

            sage: from sage.data_structures.pairing_heap import PairingHeap_of_n_integers
            sage: P = PairingHeap_of_n_integers(5)
            sage: del P
        """
        sig_free(self.nodes)

    cpdef void push(self, object item, object value) except *:
        r"""
        Insert an item into the heap with specified value (priority).

        INPUT:

        - ``item`` -- a hashable object; the item to add

        - ``value`` -- the value associated with ``item``

        EXAMPLES::

            sage: from sage.data_structures.pairing_heap import PairingHeap_of_n_hashables
            sage: P = PairingHeap_of_n_hashables(5)
            sage: P.push(1, 2)
            sage: P.top()
            (1, 2)
            sage: P.push(3, 1)
            sage: P.top()
            (3, 1)

        TESTS::

            sage: from sage.data_structures.pairing_heap import PairingHeap_of_n_hashables
            sage: P = PairingHeap_of_n_hashables(2)
            sage: P.push(1, 2)
            sage: P.push(1, 2)
            Traceback (most recent call last):
            ...
            ValueError: 1 is already in the heap
            sage: P.push(11, 2)
            sage: P.push(7, 5)
            Traceback (most recent call last):
            ...
            ValueError: the heap is full
        """
        if item in self:
            raise ValueError(f"{item} is already in the heap")
        if self.full():
            raise ValueError("the heap is full")

        cdef size_t idx = self.free_idx.pop()
        self._int_to_item[idx] = item
        self._item_to_int[item] = idx
        cdef PairingHeapNodePy * p = self.nodes + idx
        Py_INCREF(value)
        p.value = <PyObject *>value
        p.prev = p.next = p.child = NULL
        if self.root == NULL:
            self.root = p
        else:
            self.root = PairingHeapNodePy._merge(self.root, p)
        self.number_of_items += 1

    cpdef tuple top(self):
        r"""
        Return the top pair (item, value) of the heap.

        EXAMPLES::

            sage: from sage.data_structures.pairing_heap import PairingHeap_of_n_hashables
            sage: P = PairingHeap_of_n_hashables(5)
            sage: P.push(1, 2)
            sage: P.top()
            (1, 2)
            sage: P.push(3, 1)
            sage: P.top()
            (3, 1)

            sage: P = PairingHeap_of_n_hashables(3)
            sage: P.top()
            Traceback (most recent call last):
            ...
            ValueError: trying to access the top of an empty heap
        """
        if not self:
            raise ValueError("trying to access the top of an empty heap")
        cdef size_t idx = self.root - self.nodes
        return self._int_to_item[idx], <object>self.root.value

    cpdef object top_item(self):
        r"""
        Return the top item of the heap.

        EXAMPLES::

            sage: from sage.data_structures.pairing_heap import PairingHeap_of_n_hashables
            sage: P = PairingHeap_of_n_hashables(5)
            sage: P.push(1, 2)
            sage: P.top()
            (1, 2)
            sage: P.top_item()
            1

            sage: P = PairingHeap_of_n_hashables(3)
            sage: P.top_item()
            Traceback (most recent call last):
            ...
            ValueError: trying to access the top of an empty heap
        """
        if not self:
            raise ValueError("trying to access the top of an empty heap")
        cdef size_t idx = self.root - self.nodes
        return self._int_to_item[idx]

    cpdef void pop(self) noexcept:
        r"""
        Remove the top item from the heap.

        If the heap is already empty, we do nothing.

        EXAMPLES::

            sage: from sage.data_structures.pairing_heap import PairingHeap_of_n_hashables
            sage: P = PairingHeap_of_n_hashables(5); len(P)
            0
            sage: P.push(1, 2); len(P)
            1
            sage: P.push(2, 3); len(P)
            2
            sage: P.pop(); len(P)
            1
            sage: P.pop(); len(P)
            0
            sage: P.pop(); len(P)
            0
        """
        if not self:
            return
        cdef object item = self.top_item()
        cdef size_t idx = self._item_to_int[item]
        Py_XDECREF(<PyObject *>self.nodes[idx].value)
        self.free_idx.append(idx)
        del self._item_to_int[item]
        self.number_of_items -= 1
        self.root = PairingHeapNodePy._pair(self.root.child)

    cpdef void decrease(self, object item, object new_value) except *:
        r"""
        Decrease the value of specified item.

        This method is more permissive than it should as it can also be used to
        push an item in the heap.

        INPUT:

        - ``item`` -- the item to consider

        - ``new_value`` -- the new value for ``item``

        EXAMPLES::

            sage: from sage.data_structures.pairing_heap import PairingHeap_of_n_hashables
            sage: P = PairingHeap_of_n_hashables(5)
            sage: 3 in P
            False
            sage: P.decrease(3, 33)
            sage: 3 in P
            True
            sage: P.top()
            (3, 33)
            sage: P.push(1, 10)
            sage: P.top()
            (1, 10)
            sage: P.decrease(3, 7)
            sage: P.top()
            (3, 7)

        TESTS::

            sage: P = PairingHeap_of_n_hashables(5)
            sage: P.push(1, 3)
            sage: P.decrease(1, 2)
            sage: P.decrease(1, 2)
            Traceback (most recent call last):
            ...
            ValueError: the new value must be less than the current value
        """
        cdef PairingHeapNodePy * p
        cdef size_t idx
        if item in self:
            idx = self._item_to_int[item]
            p = self.nodes + idx
            if <object>p.value <= new_value:
                raise ValueError("the new value must be less than the current value")
            Py_XDECREF(<PyObject *>p.value)
            Py_INCREF(new_value)
            p.value = <PyObject *>new_value
            if p.prev != NULL:
                PairingHeapNodePy._unlink(p)
                self.root = PairingHeapNodePy._merge(self.root, p)
        else:
            self.push(item, new_value)

    def __contains__(self, object item):
        r"""
        Check whether the specified item is in the heap.

        INPUT:

        - ``item`` -- the item to consider

        EXAMPLES::

            sage: from sage.data_structures.pairing_heap import PairingHeap_of_n_hashables
            sage: P = PairingHeap_of_n_hashables(5)
            sage: 3 in P
            False
            sage: P.push(3, 33)
            sage: 3 in P
            True
            sage: 100 in P
            False
        """
        return item in self._item_to_int

    contains = __contains__

    cpdef object value(self, object item):
        r"""
        Return the value associated with the item.

        INPUT:

        - ``item`` -- the item to consider

        EXAMPLES::

            sage: from sage.data_structures.pairing_heap import PairingHeap_of_n_hashables
            sage: P = PairingHeap_of_n_hashables(5)
            sage: P.push(3, 33)
            sage: P.push(1, 10)
            sage: P.value(3)
            33
            sage: P.value(7)
            Traceback (most recent call last):
            ...
            ValueError: 7 is not in the heap
        """
        if item not in self:
            raise ValueError(f"{item} is not in the heap")
        cdef size_t idx = self._item_to_int[item]
        return <object>self.nodes[idx].value


# ==============================================================================
# Methods to check the validity of the pairing heaps
# ==============================================================================

def _test_PairingHeap_from_C(n=100):
    r"""
    Test :class:`~sage.data_structures.pairing_heap.PairingHeap`.

    INPUT:

    - ``n`` -- a strictly positive integer (default: 100); the maximum capacity
      of the heap

    TESTS::

        sage: from sage.data_structures.pairing_heap import _test_PairingHeap_from_C
        sage: _test_PairingHeap_from_C(100)
    """
    from sage.misc.prandom import randint, shuffle
    sig_on()
    cdef PairingHeap[size_t, size_t] PH = PairingHeap[size_t, size_t]()
    sig_off()

    # Initialize a list of tuples (value, item) randomly ordered
    items = list(range(n))
    values = list(range(n))
    shuffle(items)
    shuffle(values)
    cdef list Lref = list(zip(values, items))

    for value, item in Lref:
        PH.push(item, value)
        sig_check()

    L = []
    while not PH.empty():
        item, value = PH.top()
        L.append((value, item))
        PH.pop()
        sig_check()

    if L != sorted(Lref):
        raise ValueError('the order is not good')

    # Test decrease key operations. We first push items in the heap with an
    # excess of k in the value. Then we decrease the keys in a random order by
    # random values until returning to the original values. We finally check the
    # validity of the resulting ordering.
    k = 10
    dec = {item: k for item in items}
    shuffle(Lref)
    for value, item in Lref:
        PH.push(item, value + k)
        sig_check()

    L = list(items)
    while L:
        i = randint(0, len(L) - 1)
        item = L[i]
        d = randint(1, dec[item])
        dec[item] -= d
        if not dec[item]:
            L[i] = L[-1]
            L.pop()
        PH.decrease(item, PH.value(item) - d)
        sig_check()

    L = []
    while not PH.empty():
        item, value = PH.top()
        L.append((value, item))
        PH.pop()
        sig_check()

    if L != sorted(Lref):
        raise ValueError('the order is not good')

    # Different cost function
    from sage.functions.trig import sin, cos
    sig_on()
    cdef PairingHeap[size_t, pair[size_t, size_t]] HH = PairingHeap[size_t, pair[size_t, size_t]]()
    sig_off()

    for i in range(n):
        HH.push(i, (sin(i), cos(i)))
        sig_check()

    L = []
    while not HH.empty():
        L.append(HH.top())
        HH.pop()
        sig_check()

    for (u, cu), (v, cv) in zip(L, L[1:]):
        if cu > cv:
            print(u, cu, v, cv)

    # We finally show that an error is raised when trying to access the top of
    # an empty heap
    try:
        _ = HH.top()
        print("something goes wrong, the error has not been raised")
    except ValueError as msg:
        # The error has been properly handled
        pass

    try:
        _ = HH.top_item()
        print("something goes wrong, the error has not been raised")
    except ValueError as msg:
        # The error has been properly handled
        pass

    try:
        _ = HH.top_value()
        print("something goes wrong, the error has not been raised")
    except ValueError as msg:
        # The error has been properly handled
        pass

    # Or to get the value associated to an item that is not in the heap
    try:
        _ = HH.value(123)
        print("something goes wrong, the error has not been raised")
    except ValueError as msg:
        # The error has been properly handled
        pass


def _test_PairingHeap_of_n_integers(n=100):
    r"""
    Test :class:`~sage.data_structures.pairing_heap.PairingHeap_of_n_integers`.

    INPUT:

    - ``n`` -- a strictly positive integer (default: 100); the maximum capacity
      of the heap

    TESTS::

        sage: from sage.data_structures.pairing_heap import _test_PairingHeap_of_n_integers
        sage: _test_PairingHeap_of_n_integers(100)
    """
    from sage.misc.prandom import randint, shuffle

    sig_on()
    cdef PairingHeap_of_n_integers P = PairingHeap_of_n_integers(n)
    sig_off()

    # Initialize a list of tuples (value, item) randomly ordered
    cdef list items = list(range(n))
    cdef list values = list(range(n))
    shuffle(items)
    shuffle(values)
    cdef list Lref = list(zip(values, items))

    cdef int value, item
    for value, item in Lref:
        P.push(item, value)
        sig_check()

    cdef list L = []
    while P:
        item, value = P.top()
        L.append((value, item))
        P.pop()
        sig_check()

    if L != sorted(Lref):
        raise ValueError('the order is not good')

    # Test decrease key operations. We first push items in the heap with an
    # excess of k in the value. Then we decrease the keys in a random order by
    # random values until returning to the original values. We finally check the
    # validity of the resulting ordering.
    cdef int k = 10
    cdef list dec = [k] * n
    shuffle(Lref)
    for value, item in Lref:
        P.push(item, value + k)
        sig_check()

    L = list(items)
    while L:
        i = randint(0, len(L) - 1)
        item = L[i]
        d = randint(1, dec[item])
        dec[item] -= d
        if not dec[item]:
            L[i] = L[-1]
            L.pop()
        P.decrease(item, P.value(item) - d)
        sig_check()

    L = []
    while P:
        item, value = P.top()
        L.append((value, item))
        P.pop()
        sig_check()

    if L != sorted(Lref):
        raise ValueError('the order is not good')

    # We finally show that an error is raised when trying to access the top of
    # an empty heap
    try:
        _ = P.top()
        print("something goes wrong, the error has not been raised")
    except ValueError as msg:
        # The error has been properly handled
        pass

    try:
        _ = P.top_item()
        print("something goes wrong, the error has not been raised")
    except ValueError as msg:
        # The error has been properly handled
        pass

    try:
        _ = P.top_value()
        print("something goes wrong, the error has not been raised")
    except ValueError as msg:
        # The error has been properly handled
        pass

    # Or to get the value associated to an item that is not in the heap
    try:
        _ = P.value(123)
        print("something goes wrong, the error has not been raised")
    except ValueError as msg:
        # The error has been properly handled
        pass


def _test_PairingHeap_of_n_hashables(n=100):
    r"""
    Test :class:`~sage.data_structures.pairing_heap.PairingHeap_of_n_hashables`.

    INPUT:

    - ``n`` -- a strictly positive integer (default: 100); the maximum capacity
      of the heap

    TESTS::

        sage: from sage.data_structures.pairing_heap import _test_PairingHeap_of_n_hashables
        sage: _test_PairingHeap_of_n_hashables(100)
    """
    from sage.misc.prandom import randint, shuffle

    sig_on()
    cdef PairingHeap_of_n_hashables P = PairingHeap_of_n_hashables(n)
    sig_off()

    # Initialize a list of tuples (value, item) randomly ordered
    cdef list items = [(str(i), i) for i in range(n)]
    cdef list values = list(range(n))
    shuffle(items)
    shuffle(values)
    cdef list Lref = list(zip(values, items))

    for value, item in Lref:
        P.push(item, value)
        sig_check()

    cdef list L = []
    while P:
        item, value = P.top()
        L.append((value, item))
        P.pop()
        sig_check()

    if L != sorted(Lref):
        raise ValueError('the order is not good')

    # Test decrease key operations. We first push items in the heap with an
    # excess of k in the value. Then we decrease the keys in a random order by
    # random values until returning to the original values. We finally check the
    # validity of the resulting ordering.
    cdef int k = 10
    cdef dict dec = {item: k for item in items}
    shuffle(Lref)
    for value, item in Lref:
        P.push(item, value + k)
        sig_check()

    L = list(items)
    while L:
        i = randint(0, len(L) - 1)
        item = L[i]
        d = randint(1, dec[item])
        dec[item] -= d
        if not dec[item]:
            L[i] = L[-1]
            L.pop()
        P.decrease(item, P.value(item) - d)
        sig_check()

    L = []
    while P:
        item, value = P.top()
        L.append((value, item))
        P.pop()
        sig_check()

    if L != sorted(Lref):
        raise ValueError('the order is not good')

    # We finally show that an error is raised when trying to access the top of
    # an empty heap
    try:
        _ = P.top()
        print("something goes wrong, the error has not been raised")
    except ValueError as msg:
        # The error has been properly handled
        pass

    try:
        _ = P.top_item()
        print("something goes wrong, the error has not been raised")
    except ValueError as msg:
        # The error has been properly handled
        pass

    try:
        _ = P.top_value()
        print("something goes wrong, the error has not been raised")
    except ValueError as msg:
        # The error has been properly handled
        pass

    # Or to get the value associated to an item that is not in the heap
    try:
        _ = P.value(123)
        print("something goes wrong, the error has not been raised")
    except ValueError as msg:
        # The error has been properly handled
        pass


def _compare_heaps(n=100, verbose=False):
    r"""
    Check that the heaps behave the same.

    This method selects a list of instructions: push items in some order,
    decrease the values of the items in some order, extract all items in order.
    Then it applies the same instructions to a ``PairingHeap``, a
    :class:`PairingHeap_of_n_integers` and a
    :class:`PairingHeap_of_n_hashables`. It checks that all heaps report the
    list of items in the same order and it measures the running time.

    INPUT:

    - ``n`` -- a strictly positive integer (default: 100); the maximum capacity
      of the heap

    - ``verbose`` -- boolean (default: ``False``); whether to display
      information about the running times

    EXAMPLES::

        sage: from sage.data_structures.pairing_heap import _compare_heaps
        sage: _compare_heaps(n=100)
        sage: _compare_heaps(n=100, verbose=True)  # random
        PairingHeap_of_n_integers: 7.800000000024454e-05
        PairingHeap_of_n_hashables: 9.400000000026054e-05
        PairingHeap (C++): 6.899999999987472e-05
        sage: _compare_heaps(1000000, verbose=True)  # not tested (long time), random
        PairingHeap_of_n_integers: 1.5106779999999995
        PairingHeap_of_n_hashables: 4.998040000000001
        PairingHeap (C++): 1.7841750000000012
    """
    from sage.misc.prandom import shuffle
    from sage.misc.timing import cputime

    items = list(range(n))
    values = list(range(n))
    shuffle(items)
    shuffle(values)
    Lref = list(zip(values, items))
    k = 10
    dec_order = list(items)
    shuffle(dec_order)

    t = cputime()
    sig_on()
    cdef PairingHeap_of_n_integers P = PairingHeap_of_n_integers(n)
    sig_off()
    for value, item in Lref:
        P.push(item, value + k)
        sig_check()
    for item in dec_order:
        P.decrease(item, P.value(item) - k)
        sig_check()
    LP = []
    while P:
        LP.append(P.top())
        P.pop()
        sig_check()
    t = cputime(t)
    if verbose:
        print(f"PairingHeap_of_n_integers: {t}")

    t = cputime()
    sig_on()
    cdef PairingHeap_of_n_hashables Q = PairingHeap_of_n_hashables(n)
    sig_off()
    for value, item in Lref:
        Q.push(item, value + k)
        sig_check()
    for item in dec_order:
        Q.decrease(item, Q.value(item) - k)
        sig_check()
    LQ = []
    while Q:
        LQ.append(Q.top())
        Q.pop()
        sig_check()
    t = cputime(t)
    if verbose:
        print(f"PairingHeap_of_n_hashables: {t}")

    t = cputime()
    sig_on()
    cdef PairingHeap[size_t, size_t] PH = PairingHeap[size_t, size_t]()
    sig_off()
    for value, item in Lref:
        PH.push(item, value + k)
        sig_check()
    for item in dec_order:
        PH.decrease(item, PH.value(item) - k)
        sig_check()
    LPH = []
    while not PH.empty():
        LPH.append(PH.top())
        PH.pop()
        sig_check()
    t = cputime(t)
    if verbose:
        print(f"PairingHeap (C++): {t}")

    if LPH != LP or LP != LQ:
        print('something goes wrong')



# sage.doctest: needs sage.combinat sage.modules
r"""
An introduction to crystals
===========================

Informally, a crystal `\mathcal{B}` is an oriented graph with edges
colored in some set `I` such that, for each `i\in I`, each node `x`
has:

- at most one `i`-successor, denoted `f_i x`;

- at most one `i`-predecessor, denoted `e_i x`.

By convention, one writes `f_i x=\emptyset` and `e_i x=\emptyset` when
`x` has no successor resp. predecessor.

One may think of `\mathcal{B}` as essentially a deterministic
automaton whose dual is also deterministic; in this context, the
`f_i`'s and `e_i`'s are respectively the transition functions of the
automaton and of its dual, and `\emptyset` is the sink.

A crystal comes further endowed with a weight function
`\operatorname{wt} : \mathcal{B} \to L` which satisfies
appropriate conditions.

In combinatorial representation theory, crystals are used as
combinatorial data to model representations of Lie algebra.

Axiomatic definition
--------------------

Let `C` be a Cartan type (:class:`CartanType`) with index set `I`,
and `L` be a realization of the weight lattice of the type `C`.
Let `\alpha_i` and `\alpha^{\vee}_i` denote the simple roots and
coroots respectively.

A type `C` crystal is a non-empty set `\mathcal{B}` endowed with maps
`\operatorname{wt} : \mathcal{B} \to L`,
`e_i, f_i : \mathcal{B} \to \mathcal{B} \cup \{\emptyset\}`, and
`\varepsilon_i, \varphi_i : \mathcal{B} \to \ZZ \cup \{-\infty\}`
for `i \in I` satisfying the following properties for all `i \in I`:

- for `b, b^{\prime} \in \mathcal{B}`, we have
  `f_i b^{\prime} = b` if and only if `e_i b = b^{\prime}`;

- if `e_i b \in \mathcal{B}`, then:

  * `\operatorname{wt}(e_i b) = \operatorname{wt}(b) + \alpha_i`,
  * `\varepsilon_i(e_i b) = \varepsilon_i(b) - 1`,
  * `\varphi_i(e_i b) = \varphi_i(b) + 1`;

- if `f_i b \in \mathcal{B}`, then:

  * `\operatorname{wt}(f_i b) = \operatorname{wt}(b) - \alpha_i`,
  * `\varepsilon_i(f_i b) = \varepsilon_i(b) + 1`,
  * `\varphi_i(f_i b) = \varphi_i(b) - 1`;

- `\varphi_i(b) = \varepsilon_i(b) + \langle \alpha^{\vee}_i,
  \operatorname{wt}(b) \rangle`,

- if `\varphi_i(b) = -\infty` for `b \in \mathcal{B}`,
  then `e_i b = f_i b = \emptyset`.

Some further conditions are required to guarantee that this data
indeed models a representation of a Lie algebra. For finite simply
laced types a complete characterization is given by Stembridge's local
axioms [Ste2003]_.

EXAMPLES:

We construct the type `A_5` crystal on letters (or in representation
theoretic terms, the highest weight crystal of type `A_5`
corresponding to the highest weight `\Lambda_1`)::

    sage: C = crystals.Letters(['A',5]); C
    The crystal of letters for type ['A', 5]

It has a single highest weight element::

    sage: C.highest_weight_vectors()
    (1,)

A crystal is an enumerated set (see :class:`EnumeratedSets`); and we
can count and list its elements in the usual way::

    sage: C.cardinality()
    6
    sage: C.list()
    [1, 2, 3, 4, 5, 6]

as well as use it in for loops::

    sage: [x for x in C]
    [1, 2, 3, 4, 5, 6]

Here are some more elaborate crystals (see their respective
documentations)::

    sage: Tens = crystals.TensorProduct(C, C)
    sage: Spin = crystals.Spins(['B', 3])
    sage: Tab  = crystals.Tableaux(['A', 3], shape = [2,1,1])
    sage: Fast = crystals.FastRankTwo(['B', 2], shape = [3/2, 1/2])
    sage: KR = crystals.KirillovReshetikhin(['A',2,1],1,1)

One can get (currently) crude plotting via::

    sage: Tab.plot()                                                                    # needs sage.plot
    Graphics object consisting of 52 graphics primitives

If dot2tex is installed, one can obtain nice latex pictures via::

    sage: K = crystals.KirillovReshetikhin(['A',3,1], 1,1)
    sage: view(K, pdflatex=True) # optional - dot2tex graphviz, not tested (opens external window)

or with colored edges::

    sage: K = crystals.KirillovReshetikhin(['A',3,1], 1,1)
    sage: G = K.digraph()
    sage: G.set_latex_options(color_by_label={0:"black", 1:"red", 2:"blue", 3:"green"})
    sage: view(G, pdflatex=True) # optional - dot2tex graphviz, not tested (opens external window)

For rank two crystals, there is an alternative method of getting
metapost pictures. For more information see ``C.metapost?``.

.. SEEALSO:: :ref:`The overview of crystal features in Sage<sage.combinat.crystals.all>`

.. TODO::

    -  Vocabulary and conventions:

       -  For a classical crystal: connected / highest weight /
          irreducible

       -  ...

    -  Layout instructions for plot() for rank 2 types

    -  RestrictionOfCrystal


The crystals library in Sage grew up from an initial implementation in
MuPAD-Combinat (see <MuPAD-Combinat>/lib/COMBINAT/crystals.mu).
"""

#*****************************************************************************
#       Copyright (C) 2007 Anne Schilling <anne at math.ucdavis.edu>
#                          Nicolas Thiery <nthiery at users.sf.net>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
# ***************************************************************************
# Acknowledgment: most of the design and implementation of this
# library is heavily inspired from MuPAD-Combinat.
# ***************************************************************************

from sage.combinat.backtrack import GenericBacktracker


class CrystalBacktracker(GenericBacktracker):
    def __init__(self, crystal, index_set=None):
        r"""
        Time complexity: `O(nF)` amortized for each produced
        element, where `n` is the size of the index set, and `F` is
        the cost of computing `e` and `f` operators.

        Memory complexity: `O(D)` where `D` is the depth of the crystal.

        Principle of the algorithm:

        Let `C` be a classical crystal. It's an acyclic graph where each
        connected component has a unique element without predecessors (the
        highest weight element for this component). Let's assume for
        simplicity that `C` is irreducible (i.e. connected) with highest
        weight element `u`.

        One can define a natural spanning tree of `C` by taking
        `u` as the root of the tree, and for any other element
        `y` taking as ancestor the element `x` such that
        there is an `i`-arrow from `x` to `y` with
        `i` minimal. Then, a path from `u` to `y`
        describes the lexicographically smallest sequence
        `i_1,\dots,i_k` such that
        `(f_{i_k} \circ f_{i_1})(u)=y`.

        Morally, the iterator implemented below just does a depth first
        search walk through this spanning tree. In practice, this can be
        achieved recursively as follows: take an element `x`, and
        consider in turn each successor `y = f_i(x)`, ignoring
        those such that `y = f_j(x^{\prime})` for some `x^{\prime}` and
        `j<i` (this can be tested by computing `e_j(y)`
        for `j<i`).

        EXAMPLES::

            sage: from sage.combinat.crystals.crystals import CrystalBacktracker
            sage: C = crystals.Tableaux(['B',3],shape=[3,2,1])
            sage: CB = CrystalBacktracker(C)
            sage: len(list(CB))
            1617
            sage: CB = CrystalBacktracker(C, [1,2])
            sage: len(list(CB))
            8
        """
        GenericBacktracker.__init__(self, None, None)
        self._crystal = crystal
        if index_set is None:
            self._index_set = crystal.index_set()
        else:
            self._index_set = index_set

    def _rec(self, x, state):
        """
        Return an iterator for the (immediate) children of ``x`` in the search
        tree.

        EXAMPLES::

            sage: from sage.combinat.crystals.crystals import CrystalBacktracker
            sage: C = crystals.Letters(['A', 5])
            sage: CB = CrystalBacktracker(C)
            sage: list(CB._rec(C(1), 'n/a'))
            [(2, 'n/a', True)]
        """
        #We will signal the initial case by having a object and state
        #of None and consider it separately.
        if x is None and state is None:
            for gen in self._crystal.highest_weight_vectors():
                yield gen, "n/a", True
            return

        # Run through the children y of x
        for i in self._index_set:
            y = x.f(i)
            if y is None:
                continue
            # Ignore those which can be reached by an arrow with smaller label
            hasParent = False
            for j in self._index_set:
                if j == i:
                    break
                if y.e(j) is not None:
                    hasParent = True
                    break
            if hasParent:
                continue

            # yield y and all elements further below
            yield y, "n/a", True



r"""
Morphisms between number fields

This module provides classes to represent ring homomorphisms between number
fields (i.e. field embeddings).
"""

# ***************************************************************************
#       Copyright (C) 2007 William Stein <wstein@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ***************************************************************************

from sage.misc.cachefunc import cached_method

from sage.rings.morphism import RingHomomorphism_im_gens, RingHomomorphism
from sage.structure.sequence import Sequence
from sage.structure.richcmp import richcmp


class NumberFieldHomomorphism_im_gens(RingHomomorphism_im_gens):
    def __invert__(self):
        r"""
        Return the inverse of an isomorphism of absolute number fields.

        EXAMPLES::

            sage: x = polygen(ZZ, 'x')
            sage: K.<a> = NumberField(x^2 + 5)
            sage: tau1, tau2 = K.automorphisms(); tau1, tau2
            (Ring endomorphism of Number Field in a with defining polynomial x^2 + 5
              Defn: a |--> a,
             Ring endomorphism of Number Field in a with defining polynomial x^2 + 5
              Defn: a |--> -a)
            sage: ~tau1
            Ring endomorphism of Number Field in a with defining polynomial x^2 + 5
             Defn: a |--> a
            sage: ~tau2
            Ring endomorphism of Number Field in a with defining polynomial x^2 + 5
             Defn: a |--> -a

            sage: L.<z> = CyclotomicField(5)
            sage: tau1, tau2, tau3, tau4 = L.automorphisms()
            sage: (tau1, ~tau1)
            (Ring endomorphism of Cyclotomic Field of order 5 and degree 4
              Defn: z |--> z,
             Ring endomorphism of Cyclotomic Field of order 5 and degree 4
              Defn: z |--> z)
            sage: (tau2, ~tau2)
            (Ring endomorphism of Cyclotomic Field of order 5 and degree 4
              Defn: z |--> z^2,
             Ring endomorphism of Cyclotomic Field of order 5 and degree 4
              Defn: z |--> z^3)
            sage: (tau3, ~tau3)
            (Ring endomorphism of Cyclotomic Field of order 5 and degree 4
              Defn: z |--> z^3,
             Ring endomorphism of Cyclotomic Field of order 5 and degree 4
              Defn: z |--> z^2)

             sage: M.<w> = NumberField(x^4 - 5*x + 5)
             sage: phi = M.hom([z - z^2]); phi
             Ring morphism:
               From: Number Field in w with defining polynomial x^4 - 5*x + 5
               To:   Cyclotomic Field of order 5 and degree 4
               Defn: w |--> -z^2 + z
             sage: phi^-1
             Ring morphism:
               From: Cyclotomic Field of order 5 and degree 4
               To:   Number Field in w with defining polynomial x^4 - 5*x + 5
               Defn: z |--> 3/11*w^3 + 4/11*w^2 + 9/11*w - 14/11
        """
        K = self.domain()
        L = self.codomain()
        if K.degree() != L.degree():
            raise TypeError("Can only invert isomorphisms")
        V, V_into_K, _ = K.vector_space()
        _, _, L_into_W = L.vector_space()
        linear_inverse = ~V.hom([(L_into_W * self * V_into_K)(b)
                                 for b in V.basis()])
        return L.hom([(V_into_K * linear_inverse * L_into_W)(b)
                      for b in [L.gen()]])

    def preimage(self, y):
        r"""
        Compute a preimage of `y` in the domain, provided one exists.
        Raises a :exc:`ValueError` if `y` has no preimage.

        INPUT:

        - ``y`` -- an element of the codomain of ``self``

        OUTPUT:

        Returns the preimage of `y` in the domain, if one exists.
        Raises a :exc:`ValueError` if `y` has no preimage.

        EXAMPLES::

            sage: x = polygen(ZZ, 'x')
            sage: K.<a> = NumberField(x^2 - 7)
            sage: L.<b> = NumberField(x^4 - 7)
            sage: f = K.embeddings(L)[0]
            sage: f.preimage(3*b^2 - 12/7)
            3*a - 12/7
            sage: f.preimage(b)
            Traceback (most recent call last):
            ...
            ValueError: Element 'b' is not in the image of this homomorphism.

        ::

            sage: F.<b> = QuadraticField(23)
            sage: G.<a> = F.extension(x^3 + 5)
            sage: f = F.embeddings(G)[0]
            sage: f.preimage(a^3 + 2*b + 3)
            2*b - 2
        """
        # Throughout this method I am using the convention that self is a homomorphism from the number field K to the number field L
        # Therefore, I use the names K and L in place of domain and codomain

        # try to get the cached transformation matrix and vector space isomorphisms if they exist
        try:
            M,LtoV,VtoK = self._transformation_data
        except Exception:
            # get the identifications of K and L with vector spaces over Q
            V,VtoL,LtoV = self.codomain().absolute_vector_space()
            V,VtoK,KtoV = self.domain().absolute_vector_space()
            # construct the transformation matrix from K to L by making the columns be the image of the basis of V_K in V_L using the homomorphism
            from sage.matrix.constructor import matrix
            from sage.rings.rational_field import QQ
            M = matrix(QQ, [LtoV(self(VtoK(e))) for e in V.basis()]).transpose()
            self._transformation_data = (M,LtoV,VtoK)

        # get the coordinate vector of y, solve the linear system, pass to domain
        yvec = LtoV(y)                  # pass from a point in L to its vector space representation
        try:
            xvec = M.solve_right(yvec)      # solve the linear system, throws an exception if there is no solution
        except ValueError:
            raise ValueError("Element '{}' is not in the image of this homomorphism.".format(y))
        return VtoK(xvec)               # pass from the vector space representation of K back to a point in K


class RelativeNumberFieldHomomorphism_from_abs(RingHomomorphism):
    r"""
    A homomorphism from a relative number field to some other ring, stored as a
    homomorphism from the corresponding absolute field.
    """

    def __init__(self, parent, abs_hom):
        r"""
        EXAMPLES::

            sage: x = polygen(ZZ, 'x')
            sage: K.<a, b> = NumberField([x^3 + 2, x^2 + x + 1])
            sage: f = K.hom(-a*b - a, K); f
            Relative number field endomorphism of
             Number Field in a with defining polynomial x^3 + 2 over its base field
              Defn: a |--> (-b - 1)*a
                    b |--> b
            sage: type(f)
            <class 'sage.rings.number_field.homset.RelativeNumberFieldHomset_with_category.element_class'>
        """
        RingHomomorphism.__init__(self, parent)
        self._abs_hom = abs_hom
        K = abs_hom.domain()
        from_K, to_K = K.structure()
        self._from_K = from_K
        self._to_K = to_K

    def abs_hom(self):
        r"""
        Return the corresponding homomorphism from the absolute number field.

        EXAMPLES::

            sage: x = polygen(ZZ, 'x')
            sage: K.<a, b> = NumberField([x^3 + 2, x^2 + x + 1])
            sage: K.hom(a, K).abs_hom()
            Ring morphism:
              From: Number Field in a with defining polynomial
                    x^6 - 3*x^5 + 6*x^4 - 3*x^3 - 9*x + 9
              To:   Number Field in a with defining polynomial x^3 + 2 over its base field
              Defn: a |--> a - b
        """
        return self._abs_hom

    def _repr_type(self):
        r"""
        A short string to identify the type of this homomorphism.

        EXAMPLES::

            sage: x = polygen(ZZ, 'x')
            sage: K.<a, b> = NumberField([x^3 + 2, x^2 + x + 1])
            sage: K.hom(a, K)._repr_type()
            'Relative number field'
        """
        return "Relative number field"

    @cached_method
    def im_gens(self):
        r"""
        Return the images of the generators under this map.

        EXAMPLES::

            sage: x = polygen(ZZ, 'x')
            sage: K.<a, b> = NumberField([x^3 + 2, x^2 + x + 1])
            sage: K.hom(a, K).im_gens()
            [a, b]
        """
        D = self.domain()
        C = self.codomain()
        return Sequence([self(x) for x in D.gens()], universe=C, check=False, immutable=True)

    def _richcmp_(self, other, op):
        """
        Compare.

        EXAMPLES::

            sage: x = polygen(ZZ, 'x')
            sage: K.<a, b> = NumberField([x^2 - 2, x^2 - 3])
            sage: e, u, v, w = End(K)
            sage: all([u^2 == e, u*v == w, u != e])
            True
        """
        return richcmp(self.abs_hom(), other.abs_hom(), op)

    def _repr_defn(self):
        r"""
        Return a string describing the images of the generators under this map.

        EXAMPLES::

            sage: x = polygen(ZZ, 'x')
            sage: K.<a, b> = NumberField([x^3 + 2, x^2 + x + 1])
            sage: K.hom(a, K)._repr_defn()
            'a |--> a\nb |--> b'
        """
        D = self.domain()
        ig = self.im_gens()
        return '\n'.join('%s |--> %s' % (D.gen(i), ig[i])
                         for i in range(D.ngens()))

    def _call_(self, x):
        r"""
        Evaluate this map at the element ``x``.

        This is done by first
        converting ``x`` to an element of the absolute field and then
        evaluating ``self.abs_hom()`` on it.

        EXAMPLES::

            sage: x = polygen(ZZ, 'x')
            sage: K.<a, b> = NumberField([x^3 + 2, x^2 + x + 1])
            sage: K.hom(a*b, K)(17 + 3*a + 2*b) # indirect doctest
            3*b*a + 2*b + 17
        """
        return self._abs_hom(self._to_K(x))


class CyclotomicFieldHomomorphism_im_gens(NumberFieldHomomorphism_im_gens):
    pass



#############################################################
#
#    Sparse Vector over mpq_t (the GMP rationals)
#
#############################################################

from sage.libs.gmp.types cimport mpq_t

cdef struct mpq_vector:
    mpq_t *entries      # array of nonzero entries
    Py_ssize_t *positions   # positions of those nonzero entries, starting at 0
    Py_ssize_t degree       # the degree of this sparse vector
    Py_ssize_t num_nonzero  # the number of nonzero entries of this vector.

cdef int reallocate_mpq_vector(mpq_vector* v, Py_ssize_t num_nonzero) except -1
cdef int allocate_mpq_vector(mpq_vector* v, Py_ssize_t num_nonzero) except -1
cdef int mpq_vector_init(mpq_vector* v, Py_ssize_t degree, Py_ssize_t num_nonzero) except -1
cdef void mpq_vector_clear(mpq_vector* v) noexcept
cdef Py_ssize_t mpq_binary_search0(mpq_t* v, Py_ssize_t n, mpq_t x) noexcept
cdef Py_ssize_t mpq_binary_search(mpq_t* v, Py_ssize_t n, mpq_t x, Py_ssize_t* ins) noexcept
cdef int mpq_vector_get_entry(mpq_t ans, mpq_vector* v, Py_ssize_t n) except -1
cdef bint mpq_vector_is_entry_zero_unsafe(mpq_vector* v, Py_ssize_t n) noexcept
cdef object mpq_vector_to_list(mpq_vector* v)
cdef int mpq_vector_set_entry(mpq_vector* v, Py_ssize_t n, mpq_t x) except -1
cdef int mpq_vector_set_entry_str(mpq_vector* v, Py_ssize_t n, char *x_str) except -1
cdef int add_mpq_vector_init(mpq_vector* sum, mpq_vector* v, mpq_vector* w, mpq_t multiple) except -1
cdef int mpq_vector_scale(mpq_vector* v, mpq_t scalar) except -1
cdef int mpq_vector_scalar_multiply(mpq_vector* v, mpq_vector* w, mpq_t scalar) except -1
cdef int mpq_vector_cmp(mpq_vector* v, mpq_vector* w) noexcept



# sage.doctest: needs sage.combinat sage.modules
"""
Orthogonal symmetric functions

AUTHORS:

- Travis Scrimshaw (2013-11-10): Initial version
"""
# ****************************************************************************
#       Copyright (C) 2013 Travis Scrimshaw <tscrim at ucdavis.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.combinat.partition import Partitions
from sage.libs.lrcalc import lrcalc
from sage.misc.cachefunc import cached_method

from . import sfa


class SymmetricFunctionAlgebra_orthogonal(sfa.SymmetricFunctionAlgebra_generic):
    r"""
    The orthogonal symmetric function basis (or orthogonal basis, to be short).

    The orthogonal basis `\{ o_{\lambda} \}` where `\lambda` is taken over
    all partitions is defined by the following change of basis with the
    Schur functions:

    .. MATH::

        s_{\lambda} = \sum_{\mu} \left( \sum_{\nu \in H} c^{\lambda}_{\mu\nu}
        \right) o_{\mu}

    where `H` is the set of all partitions with even-width rows and
    `c^{\lambda}_{\mu\nu}` is the usual Littlewood-Richardson (LR)
    coefficients. By the properties of LR coefficients, this can be shown to
    be a upper unitriangular change of basis.

    .. NOTE::

        This is only a filtered basis, not a `\ZZ`-graded basis. However this
        does respect the induced `(\ZZ/2\ZZ)`-grading.

    INPUT:

    - ``Sym`` -- an instance of the ring of the symmetric functions

    REFERENCES:

    - [ChariKleber2000]_
    - [KoikeTerada1987]_
    - [ShimozonoZabrocki2006]_

    EXAMPLES:

    Here are the first few orthogonal symmetric functions, in various bases::

        sage: Sym = SymmetricFunctions(QQ)
        sage: o = Sym.o()
        sage: e = Sym.e()
        sage: h = Sym.h()
        sage: p = Sym.p()
        sage: s = Sym.s()
        sage: m = Sym.m()

        sage: p(o([1]))
        p[1]
        sage: m(o([1]))
        m[1]
        sage: e(o([1]))
        e[1]
        sage: h(o([1]))
        h[1]
        sage: s(o([1]))
        s[1]

        sage: p(o([2]))
        -p[] + 1/2*p[1, 1] + 1/2*p[2]
        sage: m(o([2]))
        -m[] + m[1, 1] + m[2]
        sage: e(o([2]))
        -e[] + e[1, 1] - e[2]
        sage: h(o([2]))
        -h[] + h[2]
        sage: s(o([2]))
        -s[] + s[2]

        sage: p(o([3]))
        -p[1] + 1/6*p[1, 1, 1] + 1/2*p[2, 1] + 1/3*p[3]
        sage: m(o([3]))
        -m[1] + m[1, 1, 1] + m[2, 1] + m[3]
        sage: e(o([3]))
        -e[1] + e[1, 1, 1] - 2*e[2, 1] + e[3]
        sage: h(o([3]))
        -h[1] + h[3]
        sage: s(o([3]))
        -s[1] + s[3]

        sage: Sym = SymmetricFunctions(ZZ)
        sage: o = Sym.o()
        sage: e = Sym.e()
        sage: h = Sym.h()
        sage: s = Sym.s()
        sage: m = Sym.m()
        sage: p = Sym.p()
        sage: m(o([4]))
        -m[1, 1] + m[1, 1, 1, 1] - m[2] + m[2, 1, 1] + m[2, 2] + m[3, 1] + m[4]
        sage: e(o([4]))
        -e[1, 1] + e[1, 1, 1, 1] + e[2] - 3*e[2, 1, 1] + e[2, 2] + 2*e[3, 1] - e[4]
        sage: h(o([4]))
        -h[2] + h[4]
        sage: s(o([4]))
        -s[2] + s[4]

    Some examples of conversions the other way::

        sage: o(h[3])
        o[1] + o[3]
        sage: o(e[3])
        o[1, 1, 1]
        sage: o(m[2,1])
        o[1] - 2*o[1, 1, 1] + o[2, 1]
        sage: o(p[3])
        o[1, 1, 1] - o[2, 1] + o[3]

    Some multiplication::

        sage: o([2]) * o([1,1])
        o[1, 1] + o[2] + o[2, 1, 1] + o[3, 1]
        sage: o([2,1,1]) * o([2])
        o[1, 1] + o[1, 1, 1, 1] + 2*o[2, 1, 1] + o[2, 2] + o[2, 2, 1, 1]
         + o[3, 1] + o[3, 1, 1, 1] + o[3, 2, 1] + o[4, 1, 1]
        sage: o([1,1]) * o([2,1])
        o[1] + o[1, 1, 1] + 2*o[2, 1] + o[2, 1, 1, 1] + o[2, 2, 1]
         + o[3] + o[3, 1, 1] + o[3, 2]

    Examples of the Hopf algebra structure::

        sage: o([1]).antipode()
        -o[1]
        sage: o([2]).antipode()
        -o[] + o[1, 1]
        sage: o([1]).coproduct()
        o[] # o[1] + o[1] # o[]
        sage: o([2]).coproduct()
        o[] # o[] + o[] # o[2] + o[1] # o[1] + o[2] # o[]
        sage: o([1]).counit()
        0
        sage: o.one().counit()
        1
    """

    def __init__(self, Sym):
        """
        Initialize ``self``.

        TESTS::

            sage: o = SymmetricFunctions(QQ).o()
            sage: TestSuite(o).run()
        """
        sfa.SymmetricFunctionAlgebra_generic.__init__(self, Sym, "orthogonal",
                                                      'o', graded=False)

        # We make a strong reference since we use it for our computations
        #   and so we can define the coercion below (only codomains have
        #   strong references)
        self._s = Sym.schur()

        # Setup the coercions
        M = self._s.module_morphism(self._s_to_o_on_basis, codomain=self,
                                    triangular='upper', unitriangular=True)
        M.register_as_coercion()
        Mi = self.module_morphism(self._o_to_s_on_basis, codomain=self._s,
                                  triangular='upper', unitriangular=True)
        Mi.register_as_coercion()

    @cached_method
    def _o_to_s_on_basis(self, lam):
        r"""
        Return the orthogonal symmetric function ``o[lam]`` expanded in
        the Schur basis, where ``lam`` is a partition.

        TESTS:

        Check that this is the inverse::

            sage: o = SymmetricFunctions(QQ).o()
            sage: s = SymmetricFunctions(QQ).s()
            sage: all(o(s(o[la])) == o[la] for i in range(5) for la in Partitions(i))
            True
            sage: all(s(o(s[la])) == s[la] for i in range(5) for la in Partitions(i))
            True
        """
        R = self.base_ring()
        n = sum(lam)
        return self._s._from_dict({ mu: R.sum( (-1)**j * lrcalc.lrcoef_unsafe(lam, mu, nu)
                                               for nu in Partitions(2*j)
                                                   if all(nu.arm_length(i,i) == nu.leg_length(i,i)+1
                                                          for i in range(nu.frobenius_rank()))
                                             )
                                    for j in range(n//2+1) # // 2 for horizontal dominoes
                                    for mu in Partitions(n-2*j) })

    @cached_method
    def _s_to_o_on_basis(self, lam):
        r"""
        Return the Schur symmetric function ``s[lam]`` expanded in
        the orthogonal basis, where ``lam`` is a partition.

        INPUT:

        - ``lam`` -- a partition

        OUTPUT: the expansion of ``s[lam]`` in the orthogonal basis ``self``

        EXAMPLES::

            sage: Sym = SymmetricFunctions(QQ)
            sage: s = Sym.schur()
            sage: o = Sym.orthogonal()
            sage: o._s_to_o_on_basis(Partition([]))
            o[]
            sage: o._s_to_o_on_basis(Partition([4,2,1]))
            o[1] + 2*o[2, 1] + o[2, 2, 1] + o[3]
             + o[3, 1, 1] + o[3, 2] + o[4, 1] + o[4, 2, 1]
            sage: s(o._s_to_o_on_basis(Partition([3,1]))) == s[3,1]
            True
        """
        R = self.base_ring()
        n = sum(lam)
        return self._from_dict({ mu: R.sum( lrcalc.lrcoef_unsafe(lam, mu, [2*x for x in nu])
                                            for nu in Partitions(j) )
                                 for j in range(n//2+1) # // 2 for horizontal dominoes
                                 for mu in Partitions(n-2*j) })



from sage.structure.sage_object cimport SageObject
from sage.matroids.set_system cimport SetSystem

cdef class Matroid(SageObject):
    cdef public _SageObject__custom_name
    cdef public _cached_info
    cdef int _stored_full_rank
    cdef int _stored_size

    # virtual methods
    cpdef frozenset groundset(self)
    cpdef int _rank(self, frozenset X) except? -1

    # internal methods, assuming verified input
    cpdef frozenset _max_independent(self, frozenset X)
    cpdef frozenset _circuit(self, frozenset X)
    cpdef frozenset _fundamental_circuit(self, frozenset B, e)
    cpdef frozenset _closure(self, frozenset X)
    cpdef int _corank(self, frozenset X) noexcept
    cpdef frozenset _max_coindependent(self, frozenset X)
    cpdef frozenset _cocircuit(self, frozenset X)
    cpdef frozenset _fundamental_cocircuit(self, frozenset B, e)
    cpdef frozenset _coclosure(self, frozenset X)
    cpdef frozenset _augment(self, frozenset X, frozenset Y)

    cpdef bint _is_independent(self, frozenset X) noexcept
    cpdef bint _is_basis(self, frozenset X) noexcept
    cpdef bint _is_circuit(self, frozenset X) noexcept
    cpdef bint _is_closed(self, frozenset X) noexcept
    cpdef bint _is_coindependent(self, frozenset X) noexcept
    cpdef bint _is_cobasis(self, frozenset X) noexcept
    cpdef bint _is_cocircuit(self, frozenset X) noexcept
    cpdef bint _is_coclosed(self, frozenset X) noexcept

    cpdef _minor(self, contractions, deletions)
    cpdef _has_minor(self, N, bint certificate=*)
    cpdef _line_length(self, F)
    cpdef _extension(self, element, hyperplanes)

    cdef inline _subset_internal(self, X):
        """
        Convert ``X`` to a ``frozenset`` and check that it is a subset
        of the groundset.

        See ``_subset`` for the corresponding Python method.
        """
        S = frozenset(X)
        if not self.groundset().issuperset(S):
            raise ValueError(f"{X!r} is not a subset of the groundset")
        return S

    cdef inline __subset_all(self, X):
        """
        If ``X`` is ``None``, return the groundset.

        Otherwise, do like ``_subset``:
        convert ``X`` to a ``frozenset`` and check that it is a subset
        of the groundset.

        See ``_subset_all`` for the corresponding Python method.
        """
        if X is None:
            return self.groundset()
        S = frozenset(X)
        if not self.groundset().issuperset(S):
            raise ValueError(f"{X!r} is not a subset of the groundset")
        return S

    # ** user-facing methods **
    cpdef size(self)

    # matroid oracle
    cpdef rank(self, X=*)
    cpdef full_rank(self)
    cpdef basis(self)
    cpdef max_independent(self, X)
    cpdef circuit(self, X=*)
    cpdef fundamental_circuit(self, B, e)
    cpdef closure(self, X)
    cpdef k_closure(self, X, k)

    cpdef augment(self, X, Y=*)

    cpdef corank(self, X=*)
    cpdef full_corank(self)
    cpdef cobasis(self)
    cpdef max_coindependent(self, X)
    cpdef cocircuit(self, X=*)
    cpdef fundamental_cocircuit(self, B, e)
    cpdef coclosure(self, X)

    cpdef loops(self)
    cpdef is_independent(self, X)
    cpdef is_dependent(self, X)
    cpdef is_basis(self, X)
    cpdef is_circuit(self, X)
    cpdef is_closed(self, X)
    cpdef is_subset_k_closed(self, X, int k)

    cpdef coloops(self)
    cpdef is_coindependent(self, X)
    cpdef is_codependent(self, X)
    cpdef is_cobasis(self, X)
    cpdef is_cocircuit(self, X)
    cpdef is_coclosed(self, X)

    # verification
    cpdef is_valid(self, certificate=*)

    # enumeration
    cpdef SetSystem circuits(self, k=*)
    cpdef SetSystem nonspanning_circuits(self)
    cpdef SetSystem cocircuits(self)
    cpdef SetSystem noncospanning_cocircuits(self)
    cpdef dict circuit_closures(self)
    cpdef dict nonspanning_circuit_closures(self)
    cpdef SetSystem bases(self)
    cpdef SetSystem independent_sets(self, long k=*)
    cdef SetSystem _independent_sets(self)
    cpdef SetSystem nonbases(self)
    cpdef SetSystem dependent_sets(self, long k)
    cpdef list _extend_flags(self, list flags)
    cpdef list _flags(self, long k)
    cpdef SetSystem flats(self, long k=*)
    cpdef SetSystem coflats(self, long k)
    cpdef SetSystem hyperplanes(self)
    cpdef list f_vector(self)
    cpdef list whitney_numbers(self)
    cpdef list whitney_numbers2(self)
    cpdef SetSystem broken_circuits(self, ordering=*)
    cpdef SetSystem no_broken_circuits_sets(self, ordering=*)

    # polytopes
    cpdef matroid_polytope(self)
    cpdef independence_matroid_polytope(self)

    # isomorphism
    cpdef is_isomorphic(self, other, certificate=*)
    cpdef _is_isomorphic(self, other, certificate=*)
    cpdef isomorphism(self, other)
    cpdef _isomorphism(self, other)
    cpdef equals(self, other)
    cpdef is_isomorphism(self, other, morphism)
    cpdef _is_isomorphism(self, other, morphism)
    cpdef _relabel_map(self, mapping)

    # minors, dual, truncation
    cpdef minor(self, contractions=*, deletions=*)
    cpdef contract(self, X)
    cpdef delete(self, X)
    cpdef dual(self)
    cpdef truncation(self)
    cpdef has_minor(self, N, bint certificate=*)
    cpdef has_line_minor(self, k, hyperlines=*, certificate=*)
    cpdef _has_line_minor(self, k, hyperlines, certificate=*)

    # extension
    cpdef extension(self, element=*, subsets=*)
    cpdef coextension(self, element=*, subsets=*)
    cpdef modular_cut(self, subsets)
    cpdef linear_subclasses(self, line_length=*, subsets=*)
    cpdef extensions(self, element=*, line_length=*, subsets=*)
    cpdef coextensions(self, element=*, coline_length=*, subsets=*)

    # connectivity
    cpdef simplify(self)
    cpdef cosimplify(self)
    cpdef is_simple(self)
    cpdef is_cosimple(self)
    cpdef components(self)
    cpdef is_connected(self, certificate=*)
    cpdef connectivity(self, S, T=*)
    cpdef _connectivity(self, S, T)
    cpdef is_kconnected(self, k, certificate=*)
    cpdef link(self, S, T)
    cpdef _link(self, S, T)
    cpdef _is_3connected_shifting(self, certificate=*)
    cpdef _is_4connected_shifting(self, certificate=*)
    cpdef _shifting_all(self, X, P_rows, P_cols, Q_rows, Q_cols, m)
    cpdef _shifting(self, X, X_1, Y_2, X_2, Y_1, m)
    cpdef is_3connected(self, certificate=*, algorithm=*)
    cpdef is_4connected(self, certificate=*, algorithm=*)
    cpdef _is_3connected_CE(self, certificate=*)
    cpdef _is_3connected_BC(self, certificate=*)
    cpdef _is_3connected_BC_recursion(self, basis, fund_cocircuits)
    cpdef bint is_paving(self) noexcept
    cpdef bint is_sparse_paving(self) noexcept
    cpdef girth(self)

    # representability
    cpdef _local_binary_matroid(self, basis=*)
    cpdef binary_matroid(self, randomized_tests=*, verify=*)
    cpdef is_binary(self, randomized_tests=*)
    cpdef _local_ternary_matroid(self, basis=*)
    cpdef ternary_matroid(self, randomized_tests=*, verify=*)
    cpdef is_ternary(self, randomized_tests=*)
    cpdef bint is_regular(self) noexcept
    cpdef bint is_graphic(self) noexcept

    # matroid k-closed
    cpdef is_k_closed(self, int k)

    # matroid chordality
    cpdef _is_circuit_chordal(self, frozenset C, bint certificate=*)
    cpdef is_circuit_chordal(self, C, bint certificate=*)
    cpdef is_chordal(self, k1=*, k2=*, bint certificate=*)
    cpdef chordality(self)

    # optimization
    cpdef max_weight_independent(self, X=*, weights=*)
    cpdef max_weight_coindependent(self, X=*, weights=*)
    cpdef is_max_weight_independent_generic(self, X=*, weights=*)
    cpdef is_max_weight_coindependent_generic(self, X=*, weights=*)
    cpdef intersection(self, other, weights=*)
    cpdef _intersection(self, other, weights)
    cpdef _intersection_augmentation(self, other, weights, Y)
    cpdef intersection_unweighted(self, other)
    cpdef _intersection_unweighted(self, other)
    cpdef _intersection_augmentation_unweighted(self, other, Y)
    cpdef partition(self)

    # invariants
    cpdef _internal(self, B)
    cpdef _external(self, B)
    cpdef tutte_polynomial(self, x=*, y=*)
    cpdef characteristic_polynomial(self, la=*)
    cpdef flat_cover(self, solver=*, verbose=*, integrality_tolerance=*)

    # misc
    cpdef automorphism_group(self)
    cpdef bergman_complex(self)
    cpdef augmented_bergman_complex(self)
    cpdef broken_circuit_complex(self, ordering=*)

    # visualization
    cpdef plot(self, B=*, lineorders=*, pos_method=*, pos_dict=*, save_pos=*)
    cpdef show(self, B=*, lineorders=*, pos_method=*, pos_dict=*, save_pos=*, lims=*)
    cpdef _fix_positions(self, pos_dict=*, lineorders=*)

    # construction
    cpdef direct_sum(self, matroids)
